import re
from unittest.mock import AsyncMock, Mock, PropertyMock, call

import pytest

from upsies import __homepage__, __project_name__, errors, utils
from upsies.trackers import uhd


@pytest.fixture
def tracker():
    tracker = Mock()
    tracker.name = 'uhd'
    return tracker


@pytest.fixture
def imghost():
    class MockImageHost(utils.imghosts.base.ImageHostBase):
        name = 'mock image host'
        default_config = {}
        _upload_image = AsyncMock()

    return MockImageHost()


@pytest.fixture
def uhd_tracker_jobs(imghost, tracker, tmp_path, mocker):
    content_path = tmp_path / 'Foo 2000 1080p BluRay x264-ASDF'

    uhd_tracker_jobs = uhd.UhdTrackerJobs(
        content_path=str(content_path),
        tracker=tracker,
        image_hosts=(imghost,),
        btclient=Mock(),
        torrent_destination=str(tmp_path / 'destination'),
        common_job_args={
            'home_directory': str(tmp_path / 'home_directory'),
            'ignore_cache': True,
        },
        options=None,
    )

    return uhd_tracker_jobs


@pytest.fixture
def mock_job_attributes(uhd_tracker_jobs, mocker):
    job_attrs = (
        # Interactive jobs
        'imdb_job',
        'type_job',
        'year_job',
        'quality_job',
        'version_job',
        'source_job',
        'codec_job',
        'hdr_format_job',
        'tags_job',
        'poster_job',
        'trailer_job',
        'season_job',
        'automerge_group_job',
        'scene_check_job',

        # Background jobs
        'create_torrent_job',
        'mediainfo_job',
        'screenshots_job',
        'upload_screenshots_job',
        'description_job',
    )
    for job_attr in job_attrs:
        mocker.patch.object(
            type(uhd_tracker_jobs),
            job_attr,
            PropertyMock(return_value=Mock(attr=job_attr, prejobs=())),
        )


def test_jobs_before_upload_items(uhd_tracker_jobs, mock_job_attributes, mocker):
    assert tuple(job.attr for job in uhd_tracker_jobs.jobs_before_upload) == (
        # Interactive jobs
        'imdb_job',
        'type_job',
        'year_job',
        'quality_job',
        'version_job',
        'source_job',
        'codec_job',
        'hdr_format_job',
        'tags_job',
        'poster_job',
        'trailer_job',
        'season_job',
        'automerge_group_job',
        'scene_check_job',

        # Background jobs
        'create_torrent_job',
        'mediainfo_job',
        'screenshots_job',
        'upload_screenshots_job',
        'description_job',
    )
    assert uhd_tracker_jobs.poster_job.prejobs == (
        uhd_tracker_jobs.imdb_job,
    )


def test_isolated_jobs__only_description(uhd_tracker_jobs, mock_job_attributes, mocker):
    mocker.patch.object(type(uhd_tracker_jobs), 'options', PropertyMock(return_value={'only_description': True}))
    mocker.patch.object(uhd_tracker_jobs, 'get_job_and_dependencies')
    assert uhd_tracker_jobs.isolated_jobs is uhd_tracker_jobs.get_job_and_dependencies.return_value
    assert uhd_tracker_jobs.get_job_and_dependencies.call_args_list == [
        call(uhd_tracker_jobs.description_job, uhd_tracker_jobs.screenshots_job)
    ]

def test_isolated_jobs__no_isolated_jobs(uhd_tracker_jobs, mock_job_attributes, mocker):
    mocker.patch.object(type(uhd_tracker_jobs), 'options', PropertyMock(return_value={}))
    mocker.patch.object(uhd_tracker_jobs, 'get_job_and_dependencies')
    assert uhd_tracker_jobs.isolated_jobs == ()
    assert uhd_tracker_jobs.get_job_and_dependencies.call_args_list == []


def test_type_job(uhd_tracker_jobs, mocker):
    mocker.patch.object(type(uhd_tracker_jobs), 'imdb_job', PropertyMock())
    ChoiceJob_mock = mocker.patch('upsies.jobs.dialog.ChoiceJob')
    mocker.patch.object(uhd_tracker_jobs, 'get_job_name')
    mocker.patch.object(uhd_tracker_jobs, 'make_precondition')
    mocker.patch.object(uhd_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert uhd_tracker_jobs.type_job is ChoiceJob_mock.return_value
    assert ChoiceJob_mock.call_args_list == [call(
        name=uhd_tracker_jobs.get_job_name.return_value,
        label='Type',
        precondition=uhd_tracker_jobs.make_precondition.return_value,
        prejobs=(
            uhd_tracker_jobs.imdb_job,
        ),
        autodetect=uhd_tracker_jobs.autodetect_type,
        autofinish=True,
        options=(
            ('Movie', '0'),
            ('TV', '2'),
        ),
        common_job_arg='common job argument',
    )]
    assert uhd_tracker_jobs.get_job_name.call_args_list == [call('type')]
    assert uhd_tracker_jobs.make_precondition.call_args_list == [call('type_job')]
    assert uhd_tracker_jobs.common_job_args.call_args_list == [call(ignore_cache=True)]


@pytest.mark.parametrize(
    argnames='imdb_job, imdb, exp_return_value, exp_type_calls',
    argvalues=(
        (
            Mock(is_finished=True, output=()),
            Mock(type=AsyncMock(return_value=utils.release.ReleaseType.movie)),
            None,
            [],
        ),
        (
            Mock(output=('tt123456',)),
            Mock(type=AsyncMock(return_value=utils.release.ReleaseType.movie)),
            'Movie',
            [call('tt123456')],
        ),
        (
            Mock(output=('tt123456',)),
            Mock(type=AsyncMock(return_value=utils.release.ReleaseType.season)),
            'TV',
            [call('tt123456')],
        ),
        (
            Mock(output=('tt123456',)),
            Mock(type=AsyncMock(return_value=utils.release.ReleaseType.episode)),
            'TV',
            [call('tt123456')],
        ),
        (
            Mock(output=('tt123456',)),
            Mock(type=AsyncMock(return_value=utils.release.ReleaseType.unknown)),
            None,
            [call('tt123456')],
        ),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_autodetect_type(
        imdb_job, imdb,
        exp_return_value, exp_type_calls,
        uhd_tracker_jobs, mocker,
):
    mocker.patch.object(type(uhd_tracker_jobs), 'imdb_job', PropertyMock(return_value=imdb_job))
    mocker.patch.object(type(uhd_tracker_jobs), 'imdb', PropertyMock(return_value=imdb))

    return_value = await uhd_tracker_jobs.autodetect_type('job_')
    assert return_value == exp_return_value
    assert uhd_tracker_jobs.imdb.type.call_args_list == exp_type_calls


@pytest.mark.parametrize(
    argnames='type_job, exp_return_value',
    argvalues=(
        (
            Mock(is_finished=False, output=('Movie',)),
            None,
        ),
        (
            Mock(is_finished=True, output=()),
            None,
        ),
        (
            Mock(is_finished=True, output=('Movie',)),
            utils.release.ReleaseType.movie,
        ),
        (
            Mock(is_finished=True, output=('TV',)),
            utils.release.ReleaseType.season,
        ),
    ),
    ids=('Undetermined', 'Undetermined', 'Movie', 'TV'),
)
def test_user_confirmed_type(
        type_job, exp_return_value,
        uhd_tracker_jobs, mocker,
):
    mocker.patch.object(type(uhd_tracker_jobs), 'type_job', PropertyMock(return_value=type_job))
    return_value = uhd_tracker_jobs.user_confirmed_type
    assert return_value == exp_return_value


def test_year_job(uhd_tracker_jobs, mocker):
    mocker.patch.object(type(uhd_tracker_jobs), 'imdb_job', PropertyMock())
    TextFieldJob_mock = mocker.patch('upsies.jobs.dialog.TextFieldJob')
    mocker.patch.object(uhd_tracker_jobs, 'get_job_name')
    mocker.patch.object(uhd_tracker_jobs, 'make_precondition')
    mocker.patch.object(uhd_tracker_jobs, 'get_job_and_dependencies')
    mocker.patch.object(uhd_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert uhd_tracker_jobs.year_job is TextFieldJob_mock.return_value
    assert TextFieldJob_mock.call_args_list == [call(
        name=uhd_tracker_jobs.get_job_name.return_value,
        label='Year',
        precondition=uhd_tracker_jobs.make_precondition.return_value,
        prejobs=(
            uhd_tracker_jobs.imdb_job,
        ),
        text=uhd_tracker_jobs.autodetect_year,
        nonfatal_exceptions=(
            errors.RequestError,
        ),
        normalizer=uhd_tracker_jobs.normalize_year,
        validator=uhd_tracker_jobs.validate_year,
        finish_on_success=True,
        common_job_arg='common job argument',
    )]
    assert uhd_tracker_jobs.get_job_name.call_args_list == [call('year')]
    assert uhd_tracker_jobs.make_precondition.call_args_list == [call('year_job')]
    assert uhd_tracker_jobs.common_job_args.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames=(
        'imdb_id, uhd_info, imdb_year,'
        'exp_return_value, exp_mock_calls,'
    ),
    argvalues=(
        (
            'tt123456',
            {'year': '2012'},
            '2013',
            '2012',
            [
                call.get_uhd_info('tt123456'),
            ],
        ),
        (
            'tt123456',
            {'year': ''},
            '2013',
            '2013',
            [
                call.get_uhd_info('tt123456'),
                call.imdb_year('tt123456'),
            ],
        ),
        (
            'tt123456',
            {},
            '2013',
            '2013',
            [
                call.get_uhd_info('tt123456'),
                call.imdb_year('tt123456'),
            ],
        ),
        (
            'tt123456',
            {},
            '',
            None,
            [
                call.get_uhd_info('tt123456'),
                call.imdb_year('tt123456'),
            ],
        ),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_autodetect_year(
        imdb_id, uhd_info, imdb_year,
        exp_return_value, exp_mock_calls,
        uhd_tracker_jobs, mocker,
):
    mocks = Mock()
    mocker.patch.object(type(uhd_tracker_jobs), 'imdb_job', PropertyMock(return_value=Mock(is_finished=True)))
    mocker.patch.object(type(uhd_tracker_jobs), 'imdb_id', PropertyMock(return_value=imdb_id)),
    mocks.attach_mock(
        mocker.patch.object(uhd_tracker_jobs.tracker, 'get_uhd_info', AsyncMock(return_value=uhd_info)),
        'get_uhd_info',
    )
    mocks.attach_mock(
        mocker.patch.object(uhd_tracker_jobs.imdb, 'year', AsyncMock(return_value=imdb_year)),
        'imdb_year',
    )

    return_value = await uhd_tracker_jobs.autodetect_year()
    assert return_value == exp_return_value
    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize(
    argnames='text, exp_return_value',
    argvalues=(
        ('123', '123'),
        (' 123  ', '123'),
        (' \n 123 \t ', '123'),
        ('\n\nfoo\t\t', 'foo'),
    ),
    ids=lambda v: str(v),
)
def test_normalize_year(text, exp_return_value, uhd_tracker_jobs):
    return_value = uhd_tracker_jobs.normalize_year(text)
    assert return_value == exp_return_value


@pytest.mark.parametrize(
    argnames='text, exp_exception',
    argvalues=(
        ('', ValueError('Year must not be empty.')),
        ('123foo', ValueError('Year must be a number.')),
        ('123.5', ValueError('Year must be a number.')),
        ('1799', ValueError('Year is not reasonable.')),
        ('2100', ValueError('Year is not reasonable.')),
        ('2012', None),
    ),
    ids=lambda v: str(v),
)
def test_validate_year(text, exp_exception, uhd_tracker_jobs):
    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            uhd_tracker_jobs.validate_year(text)
    else:
        return_value = uhd_tracker_jobs.validate_year(text)
        assert return_value is None


def test_season_job(uhd_tracker_jobs, mocker):
    mocker.patch.object(type(uhd_tracker_jobs), 'imdb_job', PropertyMock())
    mocker.patch.object(type(uhd_tracker_jobs), 'type_job', PropertyMock())
    TextFieldJob_mock = mocker.patch('upsies.jobs.dialog.TextFieldJob')
    mocker.patch.object(uhd_tracker_jobs, 'get_job_name')
    mocker.patch.object(uhd_tracker_jobs, 'make_precondition')
    mocker.patch.object(uhd_tracker_jobs, 'get_job_and_dependencies')
    mocker.patch.object(uhd_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert uhd_tracker_jobs.season_job is TextFieldJob_mock.return_value
    assert TextFieldJob_mock.call_args_list == [call(
        name=uhd_tracker_jobs.get_job_name.return_value,
        label='Season',
        precondition=uhd_tracker_jobs.make_precondition.return_value,
        prejobs=(
            uhd_tracker_jobs.imdb_job,
            uhd_tracker_jobs.type_job,
        ),
        hidden=uhd_tracker_jobs.season_job_is_hidden,
        text=uhd_tracker_jobs.autodetect_season,
        normalizer=uhd_tracker_jobs.normalize_season,
        validator=uhd_tracker_jobs.validate_season,
        finish_on_success=True,
        common_job_arg='common job argument',
    )]
    assert uhd_tracker_jobs.get_job_name.call_args_list == [call('season')]
    assert uhd_tracker_jobs.make_precondition.call_args_list == [call('season_job')]
    assert uhd_tracker_jobs.common_job_args.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='user_confirmed_type, exp_return_value',
    argvalues=(
        (None, True),
        (utils.release.ReleaseType.season, False),
        (utils.release.ReleaseType.episode, False),
        (utils.release.ReleaseType.movie, True),

    ),
    ids=lambda v: str(v),
)
def test_season_job_is_hidden(user_confirmed_type, exp_return_value, uhd_tracker_jobs, mocker):
    mocker.patch.object(type(uhd_tracker_jobs), 'user_confirmed_type', PropertyMock(return_value=user_confirmed_type))
    return_value = uhd_tracker_jobs.season_job_is_hidden()
    assert return_value is exp_return_value


@pytest.mark.parametrize(
    argnames=(
        'user_confirmed_type, release_name,'
        'exp_return_value,'
    ),
    argvalues=(
        (
            utils.release.ReleaseType.season,
            Mock(only_season=None),
            None,
        ),
        (
            utils.release.ReleaseType.season,
            Mock(only_season=''),
            None,
        ),
        (
            utils.release.ReleaseType.season,
            Mock(only_season='12'),
            '12',
        ),
        (
            utils.release.ReleaseType.movie,
            Mock(only_season='12'),
            '',
        ),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_autodetect_season(
        user_confirmed_type, release_name,
        exp_return_value,
        uhd_tracker_jobs, mocker,
):
    mocker.patch.object(type(uhd_tracker_jobs), 'user_confirmed_type', PropertyMock(return_value=user_confirmed_type))
    mocker.patch.object(type(uhd_tracker_jobs), 'release_name', PropertyMock(return_value=release_name))

    return_value = await uhd_tracker_jobs.autodetect_season()
    assert return_value == exp_return_value


@pytest.mark.parametrize(
    argnames='text, exp_return_value',
    argvalues=(
        ('3', '3'),
        (' 3  ', '3'),
        (' \n 3 \t ', '3'),
        ('\n\nfoo\t\t', 'foo'),
    ),
    ids=lambda v: str(v),
)
def test_normalize_season(text, exp_return_value, uhd_tracker_jobs):
    return_value = uhd_tracker_jobs.normalize_season(text)
    assert return_value == exp_return_value


@pytest.mark.parametrize(
    argnames='text, exp_exception',
    argvalues=(
        ('', None),
        ('123foo', ValueError('Season must be a number.')),
        ('123.5', ValueError('Season must be a number.')),
        ('101', ValueError('Season is not reasonable.')),
        ('-1', ValueError('Season is not reasonable.')),
        ('0', None),
        ('1', None),
        ('99', None),
    ),
    ids=lambda v: str(v),
)
def test_validate_season(text, exp_exception, uhd_tracker_jobs):
    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            uhd_tracker_jobs.validate_season(text)
    else:
        return_value = uhd_tracker_jobs.validate_season(text)
        assert return_value is None


@pytest.mark.parametrize(
    argnames='season_job, exp_return_value',
    argvalues=(
        (Mock(is_finished=False), None),
        (Mock(is_finished=True, output=()), None),
        (Mock(is_finished=True, output=('',)), None),
        (Mock(is_finished=True, output=('0',)), 0),
        (Mock(is_finished=True, output=('3',)), 3),
    ),
    ids=lambda v: str(v),
)
def test_season_number(season_job, exp_return_value, uhd_tracker_jobs, mocker):
    mocker.patch.object(type(uhd_tracker_jobs), 'season_job', PropertyMock(return_value=season_job))
    return_value = uhd_tracker_jobs.season_number
    assert return_value == exp_return_value


def test_quality_job(uhd_tracker_jobs, mocker):
    ChoiceJob_mock = mocker.patch('upsies.jobs.dialog.ChoiceJob')
    mocker.patch.object(uhd_tracker_jobs, 'get_job_name')
    mocker.patch.object(uhd_tracker_jobs, 'make_precondition')
    mocker.patch.object(uhd_tracker_jobs, 'get_job_and_dependencies')
    mocker.patch.object(uhd_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert uhd_tracker_jobs.quality_job is ChoiceJob_mock.return_value
    assert ChoiceJob_mock.call_args_list == [call(
        name=uhd_tracker_jobs.get_job_name.return_value,
        label='Quality',
        precondition=uhd_tracker_jobs.make_precondition.return_value,
        autodetect=uhd_tracker_jobs.autodetect_quality,
        autofinish=True,
        options=(
            ('mHD', 'mHD'),
            ('720p', '720p'),
            ('1080p', '1080p'),
            ('1080i', '1080i'),
            ('2160p', '2160p'),
            ('Other', 'Others'),
        ),
        common_job_arg='common job argument',
    )]
    assert uhd_tracker_jobs.get_job_name.call_args_list == [call('quality')]
    assert uhd_tracker_jobs.make_precondition.call_args_list == [call('quality_job')]
    assert uhd_tracker_jobs.common_job_args.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='resolution, exp_return_value',
    argvalues=(
        ('4320p', 'Other'),
        ('2160p', '2160p'),
        ('1080i', '1080i'),
        ('1080p', '1080p'),
        ('720p', '720p'),
        ('480p', 'Other'),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_autodetect_quality(
        resolution, exp_return_value,
        uhd_tracker_jobs, mocker,
):
    mocks = Mock()
    mocks.attach_mock(mocker.patch('upsies.utils.video.resolution', return_value=resolution), 'resolution')

    return_value = await uhd_tracker_jobs.autodetect_quality('job_')
    assert return_value == exp_return_value


def test_version_job(uhd_tracker_jobs, mocker):
    CustomJob_mock = mocker.patch('upsies.jobs.custom.CustomJob')
    mocker.patch.object(uhd_tracker_jobs, 'get_job_name')
    mocker.patch.object(uhd_tracker_jobs, 'make_precondition')
    mocker.patch.object(uhd_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert uhd_tracker_jobs.version_job is CustomJob_mock.return_value
    assert CustomJob_mock.call_args_list == [call(
        name=uhd_tracker_jobs.get_job_name.return_value,
        label='Version',
        precondition=uhd_tracker_jobs.make_precondition.return_value,
        worker=uhd_tracker_jobs.autodetect_version,
        no_output_is_ok=True,
        common_job_arg='common job argument',
    )]
    assert uhd_tracker_jobs.get_job_name.call_args_list == [call('version')]
    assert uhd_tracker_jobs.make_precondition.call_args_list == [call('version_job')]
    assert uhd_tracker_jobs.common_job_args.call_args_list == [call(ignore_cache=True)]


@pytest.mark.asyncio
async def test_autodetect_version(uhd_tracker_jobs, mocker):
    mocks = Mock(
        uncut=Mock(return_value=True),
        dc=Mock(return_value=False),
        extended=Mock(return_value=True),
        hybrid=Mock(return_value=False),
    )
    mocker.patch.object(uhd_tracker_jobs, 'version_map', {
        'Uncut': mocks.uncut,
        'DC': mocks.dc,
        'B&W': None,  # No autodetection
        'Extended': mocks.extended,
        'Hybrid': mocks.hybrid,
    })

    return_value = await uhd_tracker_jobs.autodetect_version('job_')
    assert return_value == {'Uncut', 'Extended'}


@pytest.mark.parametrize(
    argnames='version, release, exp_return_value',
    argvalues=(
        ("Director's Cut", Mock(edition=['foo', 'bar', 'baz']), False),
        ("Director's Cut", Mock(edition=['foo', "Director's Cut", 'baz']), True),
        ('Theatrical', Mock(edition=['foo', 'bar', 'baz']), False),
        ('Theatrical', Mock(edition=['foo', 'Theatrical Cut', 'baz']), True),
        ('Extended', Mock(edition=['foo', 'bar', 'baz']), False),
        ('Extended', Mock(edition=['foo', 'Extended Cut', 'baz']), True),
        ('IMAX', Mock(edition=['foo', 'bar', 'baz']), False),
        ('IMAX', Mock(edition=['foo', 'IMAX', 'baz']), True),
        ('Uncut', Mock(edition=['foo', 'bar', 'baz']), False),
        ('Uncut', Mock(edition=['foo', 'Uncut', 'baz']), True),
        ('TV Cut', Mock(), None),
        ('Unrated', Mock(edition=['foo', 'bar', 'baz']), False),
        ('Unrated', Mock(edition=['foo', 'Unrated', 'baz']), True),
        ('Remastered', Mock(edition=['foo', 'bar', 'baz']), False),
        ('Remastered', Mock(edition=['foo', 'Remastered', 'baz']), True),
        ('4K Remaster', Mock(edition=['foo', 'bar', 'baz']), False),
        ('4K Remaster', Mock(edition=['foo', '4k Remastered', 'baz']), True),
        ('4K Restoration', Mock(edition=['foo', 'bar', 'baz']), False),
        ('4K Restoration', Mock(edition=['foo', '4k Restored', 'baz']), True),
        ('B&W Version', Mock(), None),
        ('Criterion', Mock(edition=['foo', 'bar', 'baz']), False),
        ('Criterion', Mock(edition=['foo', 'Criterion Collection', 'baz']), True),
        ('2in1', Mock(edition=['foo', 'bar', 'baz']), False),
        ('2in1', Mock(edition=['foo', '2in1', 'baz']), True),
        ('3in1', Mock(edition=['foo', 'bar', 'baz']), False),
        ('3in1', Mock(edition=['foo', '3in1', 'baz']), True),
        ('Hybrid', Mock(source='BluRay'), False),
        ('Hybrid', Mock(source='Hybrid BluRay'), True),
        ('10-bit', Mock(path='path/to/release/content'), False),  # Tested in separate test
        ('Extras', Mock(), None),
    ),
    ids=lambda v: str(v),
)
def test_version_map(version, release, exp_return_value, uhd_tracker_jobs):
    if exp_return_value is None:
        assert uhd_tracker_jobs.version_map[version] is None
    else:
        is_version = uhd_tracker_jobs.version_map[version]
        assert is_version(release) is exp_return_value

@pytest.mark.parametrize(
    argnames='bit_depth, exp_return_value',
    argvalues=(
        (0, False),
        (8, False),
        (10, True),
        (11, False),
    ),
    ids=lambda v: str(v),
)
def test_version_map_bit_depth(bit_depth, exp_return_value, uhd_tracker_jobs, mocker):
    release = Mock(path='path/to/content')
    bit_depth_mock = mocker.patch('upsies.utils.video.bit_depth', return_value=bit_depth)
    is_version = uhd_tracker_jobs.version_map['10-bit']
    assert is_version(release) is exp_return_value
    assert bit_depth_mock.call_args_list == [call(release.path, default=None)]


def test_source_job(uhd_tracker_jobs, mocker):
    ChoiceJob_mock = mocker.patch('upsies.jobs.dialog.ChoiceJob')
    mocker.patch.object(uhd_tracker_jobs, 'get_job_name')
    mocker.patch.object(uhd_tracker_jobs, 'make_precondition')
    mocker.patch.object(uhd_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert uhd_tracker_jobs.source_job is ChoiceJob_mock.return_value
    assert ChoiceJob_mock.call_args_list == [call(
        name=uhd_tracker_jobs.get_job_name.return_value,
        label='Source',
        precondition=uhd_tracker_jobs.make_precondition.return_value,
        autodetect=uhd_tracker_jobs.autodetect_source,
        autofinish=True,
        options=(
            ('Blu-ray', 'Blu-ray'),
            ('Remux', 'Remux'),
            ('Encode', 'Encode'),
            ('WEB-DL', 'WEB-DL'),
            ('WEBRip', 'WEBRip'),
            ('HDRip', 'HDRip'),
            ('HDTV', 'HDTV'),
            ('Other', 'Others'),
        ),
        common_job_arg='common job argument',
    )]
    assert uhd_tracker_jobs.get_job_name.call_args_list == [call('source')]
    assert uhd_tracker_jobs.make_precondition.call_args_list == [call('source_job')]
    assert uhd_tracker_jobs.common_job_args.call_args_list == [call(ignore_cache=True)]


@pytest.mark.asyncio
async def test_autodetect_source(uhd_tracker_jobs, mocker):
    mocks = Mock(
        dvd=Mock(return_value=False),
        webrip=Mock(return_value=True),
        bluray=Mock(return_value=False),
        hdtv=Mock(return_value=True),
    )
    mocker.patch.object(uhd_tracker_jobs, 'source_map', {
        'DVD': mocks.dvd,
        'WEBRip': mocks.webrip,
        'BluRay': mocks.bluray,
        'HDTV': mocks.hdtv,
    })

    return_value = await uhd_tracker_jobs.autodetect_source('job_')
    assert return_value == 'WEBRip'

    mocks.webrip.return_value = False
    mocks.hdtv.return_value = False
    return_value = await uhd_tracker_jobs.autodetect_source('job_')
    assert return_value is None


@pytest.mark.parametrize(
    argnames='source, release, exp_return_value',
    argvalues=(
        ('Encode', Mock(source='foo BluRay baz'), True),
        ('Encode', Mock(source='foo BluRay baz'), True),
        ('Encode', Mock(source='foo bar baz'), False),
        ('Encode', Mock(source='foo HD-DVD baz'), True),
        ('WEB-DL', Mock(source='foo bar baz'), False),
        ('WEB-DL', Mock(source='foo WEB-DL baz'), True),
        ('WEBRip', Mock(source='foo bar baz'), False),
        ('WEBRip', Mock(source='foo WEBRip baz'), True),
    ),
    ids=lambda v: str(v),
)
def test_source_map(source, release, exp_return_value, uhd_tracker_jobs):
    is_source = uhd_tracker_jobs.source_map[source]
    assert is_source(release) is exp_return_value


def test_codec_job(uhd_tracker_jobs, mocker):
    ChoiceJob_mock = mocker.patch('upsies.jobs.dialog.ChoiceJob')
    mocker.patch.object(uhd_tracker_jobs, 'get_job_name')
    mocker.patch.object(uhd_tracker_jobs, 'make_precondition')
    mocker.patch.object(uhd_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert uhd_tracker_jobs.codec_job is ChoiceJob_mock.return_value
    assert ChoiceJob_mock.call_args_list == [call(
        name=uhd_tracker_jobs.get_job_name.return_value,
        label='Codec',
        precondition=uhd_tracker_jobs.make_precondition.return_value,
        autodetect=uhd_tracker_jobs.autodetect_codec,
        autofinish=True,
        options=(
            ('x264', 'x264'),
            ('x265', 'x265'),
            ('x266', 'x266'),
            ('H.264', 'H.264'),  # AVC aka H.264
            ('H.265', 'HEVC'),   # HEVC aka H.265
            ('AV1', 'AV1'),
            ('VC-1', 'VC-1'),
            ('MPEG-2', 'MPEG-2'),
        ),
        common_job_arg='common job argument',
    )]
    assert uhd_tracker_jobs.get_job_name.call_args_list == [call('codec')]
    assert uhd_tracker_jobs.make_precondition.call_args_list == [call('codec_job')]
    assert uhd_tracker_jobs.common_job_args.call_args_list == [call(ignore_cache=True)]


@pytest.mark.asyncio
async def test_autodetect_codec(uhd_tracker_jobs, mocker):
    mocks = Mock(
        x264=Mock(return_value=False),
        x265=Mock(return_value=True),
        h264=Mock(return_value=True),
        h265=Mock(return_value=False),
    )
    mocker.patch.object(uhd_tracker_jobs, 'codec_map', {
        'x264': mocks.x264,
        'x265': mocks.x265,
        'H.264': mocks.h264,
        'H.265': mocks.h265,
    })

    return_value = await uhd_tracker_jobs.autodetect_codec('job_')
    assert return_value == 'x265'

    mocks.x265.return_value = False
    mocks.h264.return_value = False
    return_value = await uhd_tracker_jobs.autodetect_codec('job_')
    assert return_value is None


@pytest.mark.parametrize(
    argnames='codec, release, exp_return_value',
    argvalues=(
        ('x264', Mock(video_format='foo'), False),
        ('x264', Mock(video_format='x264'), True),
        ('x265', Mock(video_format='foo'), False),
        ('x265', Mock(video_format='x265'), True),
        ('H.264', Mock(video_format='foo'), False),
        ('H.264', Mock(video_format='H.264'), True),
        ('H.265', Mock(video_format='foo'), False),
        ('H.265', Mock(video_format='H.265'), True),
    ),
    ids=lambda v: str(v),
)
def test_codec_map(codec, release, exp_return_value, uhd_tracker_jobs):
    is_codec = uhd_tracker_jobs.codec_map[codec]
    assert is_codec(release) is exp_return_value


def test_hdr_format_job(uhd_tracker_jobs, mocker):
    ChoiceJob_mock = mocker.patch('upsies.jobs.dialog.ChoiceJob')
    mocker.patch.object(uhd_tracker_jobs, 'get_job_name')
    mocker.patch.object(uhd_tracker_jobs, 'make_precondition')
    mocker.patch.object(uhd_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert uhd_tracker_jobs.hdr_format_job is ChoiceJob_mock.return_value
    assert ChoiceJob_mock.call_args_list == [call(
        name=uhd_tracker_jobs.get_job_name.return_value,
        label='HDR',
        precondition=uhd_tracker_jobs.make_precondition.return_value,
        autodetect=uhd_tracker_jobs.autodetect_hdr_format,
        autofinish=True,
        options=(
            ('No', 'No'),
            ('HDR10', 'HDR10'),
            ('HDR10+', 'HDR10+'),
            ('Dolby Vision', 'DoVi'),
        ),
        common_job_arg='common job argument',
    )]
    assert uhd_tracker_jobs.get_job_name.call_args_list == [call('hdr-format')]
    assert uhd_tracker_jobs.make_precondition.call_args_list == [call('hdr_format_job')]
    assert uhd_tracker_jobs.common_job_args.call_args_list == [call()]


@pytest.mark.asyncio
async def test_autodetect_hdr_format(uhd_tracker_jobs, mocker):
    mocks = Mock(
        no=Mock(return_value=False),
        hdr10=Mock(return_value=True),
        hdr10p=Mock(return_value=True),
        dovi=Mock(return_value=False),
    )
    mocker.patch.object(uhd_tracker_jobs, 'hdr_format_map', {
        'No': mocks.no,
        'HDR10': mocks.hdr10,
        'HDR10+': mocks.hdr10p,
        'Dolby Vision': mocks.dovi,
    })

    return_value = await uhd_tracker_jobs.autodetect_hdr_format('job_')
    assert return_value == 'HDR10'

    mocks.hdr10.return_value = False
    mocks.hdr10p.return_value = False
    return_value = await uhd_tracker_jobs.autodetect_hdr_format('job_')
    assert return_value is None


@pytest.mark.parametrize(
    argnames='hdr_format, release, exp_return_value',
    argvalues=(
        ('Dolby Vision', Mock(hdr_format='DV'), True),
        ('Dolby Vision', Mock(hdr_format='foo'), False),
        ('HDR10+', Mock(hdr_format='HDR10+'), True),
        ('HDR10+', Mock(hdr_format='foo'), False),
        ('HDR10', Mock(hdr_format='HDR10'), True),
        ('HDR10', Mock(hdr_format='foo'), False),
        ('No', Mock(hdr_format=''), True),
        ('No', Mock(hdr_format='?'), False),
    ),
    ids=lambda v: str(v),
)
def test_hdr_format_map(hdr_format, release, exp_return_value, uhd_tracker_jobs):
    is_hdr_format = uhd_tracker_jobs.hdr_format_map[hdr_format]
    assert is_hdr_format(release) is exp_return_value


def test_tags_job(uhd_tracker_jobs, mocker):
    mocker.patch.object(type(uhd_tracker_jobs), 'imdb_job', PropertyMock())
    TextFieldJob_mock = mocker.patch('upsies.jobs.dialog.TextFieldJob')
    mocker.patch.object(uhd_tracker_jobs, 'get_job_name')
    mocker.patch.object(uhd_tracker_jobs, 'make_precondition')
    mocker.patch.object(uhd_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert uhd_tracker_jobs.tags_job is TextFieldJob_mock.return_value
    assert TextFieldJob_mock.call_args_list == [call(
        name=uhd_tracker_jobs.get_job_name.return_value,
        label='Tags',
        precondition=uhd_tracker_jobs.make_precondition.return_value,
        prejobs=(
            uhd_tracker_jobs.imdb_job,
        ),
        text=uhd_tracker_jobs.autodetect_tags,
        nonfatal_exceptions=(
            errors.RequestError,
        ),
        finish_on_success=True,
        normalizer=uhd_tracker_jobs.normalize_tags,
        validator=uhd_tracker_jobs.validate_tags,
        common_job_arg='common job argument',
    )]
    assert uhd_tracker_jobs.get_job_name.call_args_list == [call('tags')]
    assert uhd_tracker_jobs.make_precondition.call_args_list == [call('tags_job')]
    assert uhd_tracker_jobs.common_job_args.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='imdb_id, uhd_info, exp_return_value',
    argvalues=(
        ('tt123456', {}, ''),
        ('tt123456', {'tag': ''}, ''),
        ('tt123456', {'tag': 'fo&eacute;,, bar, baz,'}, 'foé,, bar, baz,'),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_autodetect_tags(
        imdb_id, uhd_info,
        exp_return_value,
        uhd_tracker_jobs, mocker,
):
    mocker.patch.object(type(uhd_tracker_jobs), 'imdb_id', PropertyMock(return_value=imdb_id))
    mocker.patch.object(uhd_tracker_jobs.tracker, 'get_uhd_info', AsyncMock(return_value=uhd_info))

    return_value = await uhd_tracker_jobs.autodetect_tags()
    assert return_value == exp_return_value
    assert uhd_tracker_jobs.tracker.get_uhd_info.call_args_list == [call(uhd_tracker_jobs.imdb_id)]


@pytest.mark.parametrize(
    argnames='max_tags_length, text, exp_return_value',
    argvalues=(
        (15, '', ''),
        (15, 'foo', 'foo'),
        (15, 'foo,bar,baz,foo', 'foo\nbar\nbaz'),
        (15, ',  ,, foo ,\nbar\t, \t\nbaz ,   ', 'foo\nbar\nbaz'),
        (15, 'foo,bar,baz,this', 'foo\nbar\nbaz'),
        (15, 'foo,bar,baz,thi', 'foo\nbar\nbaz\nthi'),
    ),
    ids=lambda v: str(v),
)
def test_normalize_tags(
        max_tags_length, text, exp_return_value,
        uhd_tracker_jobs, mocker,
):
    mocker.patch.object(uhd_tracker_jobs, 'max_tags_length', max_tags_length)
    return_value = uhd_tracker_jobs.normalize_tags(text)
    assert return_value == exp_return_value


@pytest.mark.parametrize(
    argnames='min_tags_count, text, exp_exception',
    argvalues=(
        (3, '', ValueError('At least 3 tags are required.')),
        (3, 'foo', ValueError('At least 3 tags are required.')),
        (3, 'foo\nbar', ValueError('At least 3 tags are required.')),
        (3, 'foo\nbar\nbaz', None),
    ),
    ids=lambda v: str(v),
)
def test_validate_tags(
        min_tags_count, text, exp_exception,
        uhd_tracker_jobs, mocker,
):
    mocker.patch.object(uhd_tracker_jobs, 'min_tags_count', min_tags_count)
    if isinstance(exp_exception, Exception):
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            uhd_tracker_jobs.validate_tags(text)
    else:
        return_value = uhd_tracker_jobs.validate_tags(text)
        assert return_value is None


@pytest.mark.parametrize(
    argnames='uhd_info, exp_return_value',
    argvalues=(
        ({}, None),
        ({'photo': ''}, None),
        ({'photo': 'http://my/poster.jpg'}, {
            'poster': 'http://my/poster.jpg',
            'width': None,
            'height': None,
            'imghosts': (),
            'write_to': None,
        }),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_get_poster_from_tracker(
        uhd_info, exp_return_value,
        uhd_tracker_jobs, mocker,
):
    mocks = AsyncMock()
    mocker.patch.object(type(uhd_tracker_jobs), 'imdb_job', PropertyMock(return_value=Mock(wait=mocks.imdb_wait)))
    mocker.patch.object(type(uhd_tracker_jobs), 'imdb_id', PropertyMock(return_value='tt123456'))
    mocks.attach_mock(
        mocker.patch.object(uhd_tracker_jobs.tracker, 'get_uhd_info', AsyncMock(return_value=uhd_info)),
        'get_uhd_info',
    )

    return_value = await uhd_tracker_jobs.get_poster_from_tracker()
    assert return_value == exp_return_value

    assert mocks.mock_calls == [
        call.imdb_wait(),
        call.get_uhd_info(uhd_tracker_jobs.imdb_id),
    ]


def test_trailer_job(uhd_tracker_jobs, mocker):
    mocker.patch.object(type(uhd_tracker_jobs), 'imdb_job', PropertyMock())
    TextFieldJob_mock = mocker.patch('upsies.jobs.dialog.TextFieldJob')
    mocker.patch.object(uhd_tracker_jobs, 'get_job_name')
    mocker.patch.object(uhd_tracker_jobs, 'make_precondition')
    mocker.patch.object(uhd_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert uhd_tracker_jobs.trailer_job is TextFieldJob_mock.return_value
    assert TextFieldJob_mock.call_args_list == [call(
        name=uhd_tracker_jobs.get_job_name.return_value,
        label='Trailer',
        precondition=uhd_tracker_jobs.make_precondition.return_value,
        prejobs=(
            uhd_tracker_jobs.imdb_job,
        ),
        text=uhd_tracker_jobs.autodetect_trailer,
        nonfatal_exceptions=(
            errors.RequestError,
        ),
        finish_on_success=True,
        normalizer=uhd_tracker_jobs.normalize_trailer,
        validator=uhd_tracker_jobs.validate_trailer,
        common_job_arg='common job argument',
    )]
    assert uhd_tracker_jobs.get_job_name.call_args_list == [call('trailer')]
    assert uhd_tracker_jobs.make_precondition.call_args_list == [call('trailer_job')]
    assert uhd_tracker_jobs.common_job_args.call_args_list == [call(
        ignore_cache=bool('trailer' in uhd_tracker_jobs.options),
    )]


@pytest.mark.parametrize(
    argnames='options, imdb_id, uhd_info, exp_return_value, exp_mock_calls',
    argvalues=(
        (
            {'trailer': 'trailer_from_user'},
            'tt123456',
            {'trailer': 'trailer_from_uhd'},
            'trailer_from_user',
            [],
        ),
        (
            {'trailer': ''},
            'tt123456',
            {'trailer': 'trailer_from_uhd'},
            'trailer_from_uhd',
            [call.get_uhd_info('tt123456')],
        ),
        (
            {},
            'tt123456',
            {'trailer': 'trailer_from_uhd'},
            'trailer_from_uhd',
            [call.get_uhd_info('tt123456')],
        ),
        (
            {},
            'tt123456',
            {'trailer': ''},
            '',
            [call.get_uhd_info('tt123456')],
        ),
        (
            {},
            'tt123456',
            {},
            '',
            [call.get_uhd_info('tt123456')],
        ),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_autodetect_trailer(
        options, imdb_id, uhd_info,
        exp_return_value, exp_mock_calls,
        uhd_tracker_jobs, mocker,
):
    mocker.patch.object(type(uhd_tracker_jobs), 'options', PropertyMock(return_value=options))
    mocker.patch.object(type(uhd_tracker_jobs), 'imdb_job', PropertyMock(return_value=Mock(is_finished=True)))
    mocker.patch.object(type(uhd_tracker_jobs), 'imdb_id', PropertyMock(return_value=imdb_id))
    mocks = Mock()
    mocks.attach_mock(
        mocker.patch.object(uhd_tracker_jobs.tracker, 'get_uhd_info', AsyncMock(return_value=uhd_info)),
        'get_uhd_info',
    )

    return_value = await uhd_tracker_jobs.autodetect_trailer()
    assert return_value == exp_return_value
    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize(
    argnames='text, get_youtube_id, exp_return_value',
    argvalues=(
        (
            'http://youtube.com/watch?v=d34db33f',
            Mock(return_value='d34db33f'),
            'https://youtu.be/d34db33f',
        ),
        (
            'http://example.org/watch?v=d34db33f',
            Mock(side_effect=ValueError('Invalid ID or URL')),
            'http://example.org/watch?v=d34db33f',
        ),
    ),
    ids=lambda v: str(v),
)
def test_normalize_trailer(
        text, get_youtube_id,
        exp_return_value,
        uhd_tracker_jobs, mocker,
):
    mocker.patch.object(uhd_tracker_jobs, 'get_youtube_id', get_youtube_id),

    return_value = uhd_tracker_jobs.normalize_trailer(text)
    assert return_value == exp_return_value
    assert uhd_tracker_jobs.get_youtube_id.call_args_list == [
        call.get_youtube_id(text),
    ]


@pytest.mark.parametrize(
    argnames='text, get_youtube_id, exp_result, exp_mock_calls',
    argvalues=(
        (
            'http://youtu.be/d34db33f',
            Mock(return_value='d34db33f'),
            None,
            [call.get_youtube_id('http://youtu.be/d34db33f')],
        ),
        (
            'http://youtu.be/d34db33f',
            Mock(side_effect=ValueError('Invalid ID or URL')),
            ValueError('Invalid ID or URL'),
            [call.get_youtube_id('http://youtu.be/d34db33f')],
        ),
        (
            '',
            Mock(return_value='d34db33f'),
            None,
            [],
        ),
        (
            '',
            Mock(side_effect=ValueError('Invalid ID or URL')),
            None,
            [],
        ),
    ),
    ids=lambda v: str(v),
)
def test_validate_trailer(
        text, get_youtube_id,
        exp_result, exp_mock_calls,
        uhd_tracker_jobs, mocker,
):
    mocker.patch.object(uhd_tracker_jobs, 'get_youtube_id', get_youtube_id),

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            uhd_tracker_jobs.validate_trailer(text)
    else:
        return_value = uhd_tracker_jobs.validate_trailer(text)
        assert return_value == exp_result

    assert uhd_tracker_jobs.get_youtube_id.call_args_list == exp_mock_calls


@pytest.mark.parametrize('scheme', ('http', 'https'))
@pytest.mark.parametrize('host', ('www.youtube.com', 'youtube.com', 'youtu.be', 'youtube.fi'))
@pytest.mark.parametrize(
    argnames='url, default, exp_result',
    argvalues=(
        # Valid URLs
        ('{scheme}://{host}/watch?v=d34d-b33f_&feature=feedrec_grec_index', None, 'd34d-b33f_'),
        ('{scheme}://{host}/v/d34d-b33f_?fs=1&amp;hl=en_US&amp;rel=0', None, 'd34d-b33f_'),
        ('{scheme}://{host}/watch?v=d34d-b33f_#t=0m10s', None, 'd34d-b33f_'),
        ('{scheme}://{host}/embed/d34d-b33f_?rel=0', None, 'd34d-b33f_'),
        ('{scheme}://{host}/watch?v=d34d-b33f_', None, 'd34d-b33f_'),
        ('{scheme}://{host}/d34d-b33f_', None, 'd34d-b33f_'),
        ('d34d-b33f_', None, 'd34d-b33f_'),
        # Invalid URLs without default value
        ('{scheme}://{host}/', uhd.UhdTrackerJobs._YOUTUBE_ID_NO_DEFAULT, ValueError('Not a YouTube ID or URL.')),
        ('/d34d-b33f_/', uhd.UhdTrackerJobs._YOUTUBE_ID_NO_DEFAULT, ValueError('Not a YouTube ID or URL.')),
        ('http://example.org/v/d34d-b33f_', uhd.UhdTrackerJobs._YOUTUBE_ID_NO_DEFAULT, ValueError('Not a YouTube ID or URL.')),
        ('http://example.org/watch?v=d34d-b33f_', uhd.UhdTrackerJobs._YOUTUBE_ID_NO_DEFAULT, ValueError('Not a YouTube ID or URL.')),
        # Invalid URLs with default value
        ('{scheme}://{host}/', None, None),
        ('/d34d-b33f_/', '', ''),
        ('http://example.org/v/d34d-b33f_', 0, 0),
        ('http://example.org/watch?v=d34d-b33f_', 'foo', 'foo'),
    ),
    ids=lambda v: str(v),
)
def test_get_youtube_id(scheme, host, url, default, exp_result, uhd_tracker_jobs):
    url = url.format(scheme=scheme, host=host)
    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            uhd_tracker_jobs.get_youtube_id(url, default=default)
    else:
        return_value = uhd_tracker_jobs.get_youtube_id(url, default=default)
        assert return_value == exp_result


def test_description_job(uhd_tracker_jobs, mocker):
    mocker.patch.object(type(uhd_tracker_jobs), 'upload_screenshots_job', PropertyMock())
    mocker.patch.object(type(uhd_tracker_jobs), 'mediainfo_job', PropertyMock())
    TextFieldJob_mock = mocker.patch('upsies.jobs.dialog.TextFieldJob')
    mocker.patch.object(uhd_tracker_jobs, 'get_job_name')
    mocker.patch.object(uhd_tracker_jobs, 'make_precondition')
    mocker.patch.object(uhd_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert uhd_tracker_jobs.description_job is TextFieldJob_mock.return_value
    assert TextFieldJob_mock.call_args_list == [call(
        name=uhd_tracker_jobs.get_job_name.return_value,
        label='Description',
        precondition=uhd_tracker_jobs.make_precondition.return_value,
        prejobs=(
            uhd_tracker_jobs.upload_screenshots_job,
            uhd_tracker_jobs.mediainfo_job,
        ),
        text=uhd_tracker_jobs.generate_description,
        hidden=True,
        finish_on_success=True,
        read_only=True,
        common_job_arg='common job argument',
    )]
    assert uhd_tracker_jobs.get_job_name.call_args_list == [call('description')]
    assert uhd_tracker_jobs.make_precondition.call_args_list == [call('description_job')]
    assert uhd_tracker_jobs.common_job_args.call_args_list == [call(ignore_cache=True)]


def test_image_host_config(uhd_tracker_jobs):
    assert uhd_tracker_jobs.image_host_config == {
        'common': {'thumb_width': 350},
    }


@pytest.mark.parametrize('nfo', (None, '', '<nfo>'))
def test_generate_description(nfo, uhd_tracker_jobs, mocker):
    mocker.patch.object(uhd_tracker_jobs, 'generate_description_screenshots', return_value='<screenshots>')
    mocker.patch.object(uhd_tracker_jobs, 'generate_description_mediainfo', return_value='<mediainfo>')
    mocker.patch.object(uhd_tracker_jobs, 'generate_description_nfo', return_value=nfo)
    exp_promotion = (
        '[align=right][size=1]'
        f'Shared with [url={__homepage__}]{__project_name__}[/url]'
        '[/size][/align]'
    )
    return_value = uhd_tracker_jobs.generate_description()
    exp_return_value = (
        '<mediainfo>'
        + (f'\n\n{nfo}' if nfo else '')
        + '\n\n[center]<screenshots>[/center]'
        + f'\n{exp_promotion}'
    )
    assert return_value == exp_return_value
    assert uhd_tracker_jobs.generate_description_screenshots.call_args_list == [call()]
    assert uhd_tracker_jobs.generate_description_mediainfo.call_args_list == [call()]
    assert uhd_tracker_jobs.generate_description_nfo.call_args_list == [call()]


@pytest.mark.parametrize('mediainfo', ('<mediainfo>', '  <mediainfo>\n  \n\n'))
def test_generate_description_mediainfo(mediainfo, uhd_tracker_jobs, mocker):
    mocker.patch.object(type(uhd_tracker_jobs), 'mediainfo_job', PropertyMock(return_value=Mock(
        is_finished=True,
        output=(mediainfo,),
    )))

    return_value = uhd_tracker_jobs.generate_description_mediainfo()
    assert return_value == (
        '[spoiler=Mediainfo]'
        + '[code]'
        + mediainfo.strip()
        + '[/code]'
        + '[/spoiler]'
    )


def test_generate_description_screenshots(uhd_tracker_jobs, mocker):
    mocker.patch.object(type(uhd_tracker_jobs), 'upload_screenshots_job', PropertyMock(return_value=Mock(
        is_finished=True,
        uploaded_images=('this.png', 'that.png'),
    )))
    mocker.patch.object(uhd_tracker_jobs, 'make_screenshots_grid')

    return_value = uhd_tracker_jobs.generate_description_screenshots()
    assert return_value is uhd_tracker_jobs.make_screenshots_grid.return_value
    assert uhd_tracker_jobs.make_screenshots_grid.call_args_list == [
        call(
            screenshots=uhd_tracker_jobs.upload_screenshots_job.uploaded_images,
            columns=2,
            horizontal_spacer='   ',
            vertical_spacer='\n\n',
        ),
    ]


@pytest.mark.parametrize(
    argnames='nfo, exp_return_value',
    argvalues=(
        ('', None),
        ('<nfo>', '[spoiler=NFO][code]<nfo>[/code][/spoiler]'),
    ),
    ids=lambda v: repr(v),
)
def test_generate_description_nfo(nfo, exp_return_value, uhd_tracker_jobs, mocker):
    mocker.patch.object(uhd_tracker_jobs, 'read_nfo', return_value=nfo)

    return_value = uhd_tracker_jobs.generate_description_nfo()
    assert return_value == exp_return_value
    assert uhd_tracker_jobs.read_nfo.call_args_list == [call(strip=True)]


def test_automerge_group_job(uhd_tracker_jobs, mocker):
    mocker.patch.object(type(uhd_tracker_jobs), 'imdb_job', PropertyMock())
    mocker.patch.object(type(uhd_tracker_jobs), 'type_job', PropertyMock())
    mocker.patch.object(type(uhd_tracker_jobs), 'season_job', PropertyMock())
    ChoiceJob_mock = mocker.patch('upsies.jobs.dialog.ChoiceJob')
    mocker.patch.object(uhd_tracker_jobs, 'get_job_name')
    mocker.patch.object(uhd_tracker_jobs, 'make_precondition')
    mocker.patch.object(uhd_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert uhd_tracker_jobs.automerge_group_job is ChoiceJob_mock.return_value
    assert ChoiceJob_mock.call_args_list == [call(
        name=uhd_tracker_jobs.get_job_name.return_value,
        label='Automerge Group',
        precondition=uhd_tracker_jobs.make_precondition.return_value,
        prejobs=(
            uhd_tracker_jobs.imdb_job,
            uhd_tracker_jobs.type_job,
            uhd_tracker_jobs.season_job,
        ),
        autodetect=uhd_tracker_jobs.autodetect_automerge_group,
        autofinish=True,
        options=(
            ('Yes', True),
            ('No', False),
        ),
        common_job_arg='common job argument',
    )]
    assert uhd_tracker_jobs.get_job_name.call_args_list == [call('automerge-group')]
    assert uhd_tracker_jobs.make_precondition.call_args_list == [call('automerge_group_job')]
    assert uhd_tracker_jobs.common_job_args.call_args_list == [call(ignore_cache=True)]


@pytest.mark.parametrize(
    argnames='season_number, group_names, exp_return_value',
    argvalues=(
        (
            3,
            [],
            'No',
        ),
        (
            3,
            ['foo season 01', 'foo season 02', 'foo season 03'],
            'Yes',
        ),
        (
            3,
            ['foo season 01', 'foo season 02', 'foo season 03 / aka le foo'],
            'Yes',
        ),
        (
            3,
            ['season 01', 'season 02', 'season 03 / aka le foo'],
            'No',
        ),
        (
            4,
            ['foo season 01', 'foo season 02', 'foo season 03'],
            'No',
        ),
        (
            0,
            ['foo season 00', 'foo season 01', 'foo season 02'],
            'Yes',
        ),
        (
            0,
            ['foo season 01', 'foo season 02', 'foo season 03'],
            'No',
        ),
        (
            None,
            ['irrelevant'],
            'Yes',
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test_autodetect_automerge_group_with_season_number(
        season_number, group_names,
        exp_return_value,
        uhd_tracker_jobs, mocker,
):
    mocker.patch.object(uhd_tracker_jobs, '_get_season_group_names', AsyncMock(return_value=group_names))
    mocker.patch.object(type(uhd_tracker_jobs), 'season_number', PropertyMock(return_value=season_number))
    mocker.patch.object(type(uhd_tracker_jobs), 'season_job', PropertyMock(return_value=Mock(
        is_finished=True,
    )))

    return_value = await uhd_tracker_jobs.autodetect_automerge_group('ignored job')
    assert return_value == exp_return_value
    if season_number is not None:
        assert uhd_tracker_jobs._get_season_group_names.call_args_list == [call()]
    else:
        assert uhd_tracker_jobs._get_season_group_names.call_args_list == []


@pytest.mark.parametrize(
    argnames='html, exp_group_names',
    argvalues=(
        (
            (
                '<html><body>'
                '<a class="torrent_name"> <b> Foo Season 01 </b> </a>'
                '<a class="torrent_name"> <b> Fóǫ Season 02 </b> </a>'
                '<a class="torrent_name"> <b> Föö Season 03 </b> </a>'
                '</body></html>'
            ),
            [
                'foo season 01',
                'foo season 02',
                'foo season 03',
            ],
        ),
        (
            (
                '<html><body>'
                '<a>foo</a>'
                '<a>bar</a>'
                '<a>baz</a>'
                '</body></html>'
            ),
            [],
        ),
    ),
)
@pytest.mark.asyncio
async def test__get_season_group_names(html, exp_group_names, uhd_tracker_jobs, mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(uhd_tracker_jobs.tracker, 'login', AsyncMock()), 'login')
    mocks.attach_mock(mocker.patch.object(uhd_tracker_jobs.tracker, '_request', AsyncMock(return_value=html)), '_request')
    mocker.patch.object(type(uhd_tracker_jobs), 'imdb_id', PropertyMock(return_value='tt123456'))
    mocker.patch.object(
        type(uhd_tracker_jobs._tracker), '_torrents_url', PropertyMock(return_value='http://torrents.php'),
        create=True,
    )

    return_value = await uhd_tracker_jobs._get_season_group_names()
    assert return_value == exp_group_names
    assert mocks.mock_calls == [
        call.login(),
        call._request(
            method='GET',
            url=uhd_tracker_jobs._tracker._torrents_url,
            params={
                'searchstr': uhd_tracker_jobs.imdb_id,
            },
            error_prefix='Automerge group check failed',
        )
    ]


def test_release_name_translation(uhd_tracker_jobs):
    assert uhd_tracker_jobs.release_name_translation == {
        'group': {
            re.compile(r'^NOGROUP$'): 'Unknown',
        },
    }


@pytest.mark.parametrize('season_number', (None, 0, 3))
@pytest.mark.parametrize('is_scene, exp_scene', ((False, None), (True, '1')), ids=('!Scene', 'Scene'))
@pytest.mark.parametrize('automerge_group, exp_auto_merge_group', ((False, None), (True, 'on')), ids=('!Automerge', 'Automerge'))
@pytest.mark.parametrize('internal, exp_internal', ((False, None), (True, 'on')), ids=('!Internal', 'Internal'))
@pytest.mark.parametrize('_3d, exp_3d', ((False, None), (True, '1')), ids=('!3D', '3D'))
@pytest.mark.parametrize('vie, exp_vie', ((False, None), (True, '1')), ids=('!Vie', 'Vie'))
@pytest.mark.parametrize('anonymous, exp_anonymous', ((False, None), (True, '1')), ids=('!Anonymous', 'Anonymous'))
def test_post_data(
        season_number,
        is_scene, exp_scene,
        automerge_group, exp_auto_merge_group,
        internal, exp_internal,
        _3d, exp_3d,
        vie, exp_vie,
        anonymous, exp_anonymous,
        uhd_tracker_jobs, mock_job_attributes, mocker,
):
    job_outputs = {
        uhd_tracker_jobs.imdb_job: 'tt123456',
        uhd_tracker_jobs.year_job: '2012',
        uhd_tracker_jobs.quality_job: '720p',
        uhd_tracker_jobs.version_job: ('Uncut', 'Extended'),
        uhd_tracker_jobs.source_job: 'BluRay',
        uhd_tracker_jobs.codec_job: 'x264',
        uhd_tracker_jobs.hdr_format_job: 'HDR10',
        uhd_tracker_jobs.tags_job: 'drama\ncomedy\nhorror',
        uhd_tracker_jobs.poster_job: 'http://my.poster.jpg',
        uhd_tracker_jobs.trailer_job: 'https://youtu.be/MnT0nGmFkJ-IqOe',
        uhd_tracker_jobs.mediainfo_job: '[mediainfo]...[/mediainfo]',
        uhd_tracker_jobs.description_job: '[description]...[/description]',
    }

    def get_job_output(job, **__):
        if job is not None:
            print('!!!', job_outputs[job])
            return job_outputs[job]

    mocker.patch.object(uhd_tracker_jobs, 'get_job_output', side_effect=get_job_output)

    job_attributes = {
        uhd_tracker_jobs.type_job: {'choice': '0 or 2'},
        uhd_tracker_jobs.automerge_group_job: {'choice': automerge_group},
        uhd_tracker_jobs.scene_check_job: {'is_scene_release': is_scene},
    }

    def get_job_attribute(job, attribute):
        if job is not None:
            return job_attributes[job][attribute]

    mocker.patch.object(uhd_tracker_jobs, 'get_job_attribute', side_effect=get_job_attribute)

    mocker.patch.object(type(uhd_tracker_jobs), 'options', PropertyMock(return_value={
        'internal': internal,
        '3d': _3d,
        'vie': vie,
        'anonymous': anonymous,
    }))
    mocker.patch.object(type(uhd_tracker_jobs), 'release_name', PropertyMock(return_value=Mock(
        title='Original Title',
        title_aka='Also Known As',
        group='ASDF',
    )))
    mocker.patch.object(type(uhd_tracker_jobs), 'season_number', PropertyMock(return_value=season_number))

    assert uhd_tracker_jobs.post_data == {
        'submit': 'true',
        'type': job_attributes[uhd_tracker_jobs.type_job]['choice'],
        'imdbid': job_outputs[uhd_tracker_jobs.imdb_job],
        'title': uhd_tracker_jobs.release_name.title,
        'OtherTitle': uhd_tracker_jobs.release_name.title_aka,
        'smalldesc': '',
        'year': job_outputs[uhd_tracker_jobs.year_job],
        'season': season_number,
        'format': job_outputs[uhd_tracker_jobs.quality_job],
        'team': uhd_tracker_jobs.release_name.group,
        'Version': ' / '.join(job_outputs[uhd_tracker_jobs.version_job]),
        'media': job_outputs[uhd_tracker_jobs.source_job],
        'codec': job_outputs[uhd_tracker_jobs.codec_job],
        'hdr': job_outputs[uhd_tracker_jobs.hdr_format_job],
        'genre_tags': '---',
        'tags': ','.join(job_outputs[uhd_tracker_jobs.tags_job].split('\n')),
        'image': job_outputs[uhd_tracker_jobs.poster_job],
        'trailer': 'MnT0nGmFkJ-IqOe',
        'mediainfo': job_outputs[uhd_tracker_jobs.mediainfo_job],
        'release_desc': job_outputs[uhd_tracker_jobs.description_job],
        'auto_merge_group': exp_auto_merge_group,
        'internal': exp_internal,
        'd3d': exp_3d,
        'vie': exp_vie,
        'scene': exp_scene,
        'anonymous': exp_anonymous,
    }

    exp_get_job_output_calls = [
        call(uhd_tracker_jobs.imdb_job, slice=0),
        call(uhd_tracker_jobs.year_job, slice=0),
        call(uhd_tracker_jobs.quality_job, slice=0),
        call(uhd_tracker_jobs.version_job),
        call(uhd_tracker_jobs.source_job, slice=0),
        call(uhd_tracker_jobs.codec_job, slice=0),
        call(uhd_tracker_jobs.hdr_format_job, slice=0),
        call(uhd_tracker_jobs.tags_job, slice=0),
        call(uhd_tracker_jobs.poster_job, slice=0),
        call(uhd_tracker_jobs.trailer_job, slice=0),
        call(uhd_tracker_jobs.mediainfo_job, slice=0),
        call(uhd_tracker_jobs.description_job, slice=0),
    ]
    assert uhd_tracker_jobs.get_job_output.call_args_list == exp_get_job_output_calls

    assert uhd_tracker_jobs.get_job_attribute.call_args_list == [
        call(uhd_tracker_jobs.type_job, 'choice'),
        call(uhd_tracker_jobs.automerge_group_job, 'choice'),
        call(uhd_tracker_jobs.scene_check_job, 'is_scene_release'),
    ]


def test_post_files(uhd_tracker_jobs, mocker):
    mocker.patch.object(type(uhd_tracker_jobs), 'torrent_filepath', PropertyMock(return_value='path/to/file.torrent'))

    assert uhd_tracker_jobs.post_files == {
        'file_input': {
            'file': 'path/to/file.torrent',
            'mimetype': 'application/x-bittorrent',
        },
    }



import hashlib
import re
from unittest.mock import AsyncMock, Mock, PropertyMock, call

import pytest

from upsies import errors
from upsies.jobs.poster import PosterJob


@pytest.fixture
def job(tmp_path):
    return PosterJob(
        getter=AsyncMock(),
        cache_directory=str(tmp_path / 'cache'),
        home_directory=str(tmp_path / 'home'),
    )


def test_cache_id(job):
    assert job.cache_id is None


def test_initialize(mocker):
    mocks = Mock()
    mocker.patch('upsies.jobs.poster.PosterJob.signal', PropertyMock())
    job = PosterJob(
        getter=mocks.getter,
        width=mocks.width,
        height=mocks.height,
        write_to=mocks.write_to,
        imghosts=mocks.imghosts,
    )

    assert job._getter is mocks.getter
    assert job._width is mocks.width
    assert job._height is mocks.height
    assert job._write_to is mocks.write_to
    assert job._imghosts is mocks.imghosts

    assert job.signal.add.call_args_list == [
        call('obtaining'),
        call('obtained'),
        call('downloading'),
        call('downloaded'),
        call('resizing'),
        call('resized'),
        call('uploading'),
        call('uploaded'),
    ]


@pytest.mark.parametrize(
    argnames=(
        'width, height, write_to, imghosts,'
        'obtain, resize, write, upload, get_poster_filename,'
        'exp_mock_calls,'
    ),
    argvalues=(
        pytest.param(
            123, 456, 'out.jpg', ('imghost1', 'imghost2'),
            AsyncMock(return_value={
                'poster': 'poster.jpg',
                'width': 456,
                'height': 789,
                'write_to': 'farout.jpg',
                'imghosts': ('imghost4', 'imghost3'),
            }),
            AsyncMock(return_value='maybe_resized.jpg'),
            AsyncMock(),
            AsyncMock(),
            Mock(),
            [
                call._obtain(),
                call._resize('poster.jpg', 456, 789),
                call._write('maybe_resized.jpg', 'farout.jpg'),
                call._upload('maybe_resized.jpg', ('imghost4', 'imghost3')),
            ],
            id='Parameters from getter',
        ),

        pytest.param(
            123, 456, 'out.jpg', ('imghost1', 'imghost2'),
            AsyncMock(return_value={'poster': 'poster.jpg'}),
            AsyncMock(return_value='maybe_resized.jpg'),
            AsyncMock(),
            AsyncMock(),
            Mock(),
            [
                call._obtain(),
                call._resize('poster.jpg', 123, 456),
                call._write('maybe_resized.jpg', 'out.jpg'),
                call._upload('maybe_resized.jpg', ('imghost1', 'imghost2')),
            ],
            id='Parameters from initialization',
        ),

        pytest.param(
            123, None, None, (),
            AsyncMock(return_value={'poster': 'poster.jpg'}),
            AsyncMock(return_value='maybe_resized.jpg'),
            AsyncMock(),
            AsyncMock(),
            Mock(return_value='out_forced.jpg'),
            [
                call._obtain(),
                call._resize('poster.jpg', 123, None),
                call._write('maybe_resized.jpg', None),
                call._upload('maybe_resized.jpg', ()),
                call._get_poster_filename('maybe_resized.jpg'),
                call._write('maybe_resized.jpg', 'out_forced.jpg'),
            ],
            id='No output file or image host specified, but width',
        ),

        pytest.param(
            None, 456, None, (),
            AsyncMock(return_value={'poster': 'poster.jpg'}),
            AsyncMock(return_value='maybe_resized.jpg'),
            AsyncMock(),
            AsyncMock(),
            Mock(return_value='out_forced.jpg'),
            [
                call._obtain(),
                call._resize('poster.jpg', None, 456),
                call._write('maybe_resized.jpg', None),
                call._upload('maybe_resized.jpg', ()),
                call._get_poster_filename('maybe_resized.jpg'),
                call._write('maybe_resized.jpg', 'out_forced.jpg'),
            ],
            id='No output file or image host specified, but height',
        ),

        pytest.param(
            None, None, None, (),
            AsyncMock(return_value={'poster': 'poster.jpg'}),
            AsyncMock(return_value='maybe_resized.jpg'),
            AsyncMock(),
            AsyncMock(),
            Mock(return_value='out_forced.jpg'),
            [
                call._obtain(),
                call._resize('poster.jpg', None, None),
                call._write('maybe_resized.jpg', None),
                call._upload('maybe_resized.jpg', ()),
                call.send('maybe_resized.jpg'),
            ],
            id='No output file or image host specified, no resizing',
        ),

        pytest.param(
            None, None, '', (),
            AsyncMock(side_effect=PosterJob._ProcessingError('failed to obtain')),
            AsyncMock(return_value='maybe_resized.jpg'),
            AsyncMock(),
            AsyncMock(),
            Mock(),
            [
                call._obtain(),
                call.error(PosterJob._ProcessingError('failed to obtain')),
            ],
            id='Exception from _obtain()',
        ),

        pytest.param(
            None, None, '', (),
            AsyncMock(return_value={'poster': 'poster.jpg'}),
            AsyncMock(side_effect=PosterJob._ProcessingError('failed to resize')),
            AsyncMock(),
            AsyncMock(),
            Mock(),
            [
                call._obtain(),
                call._resize('poster.jpg', None, None),
                call.error(PosterJob._ProcessingError('failed to resize')),
            ],
            id='Exception from _resize()',
        ),

        pytest.param(
            None, None, '', (),
            AsyncMock(return_value={'poster': 'poster.jpg'}),
            AsyncMock(return_value='maybe_resized.jpg'),
            AsyncMock(side_effect=PosterJob._ProcessingError('failed to write')),
            AsyncMock(),
            Mock(),
            [
                call._obtain(),
                call._resize('poster.jpg', None, None),
                call._write('maybe_resized.jpg', ''),
                call.error(PosterJob._ProcessingError('failed to write')),
            ],
            id='Exception from _write()',
        ),

        pytest.param(
            None, None, '', (),
            AsyncMock(return_value={'poster': 'poster.jpg'}),
            AsyncMock(return_value='maybe_resized.jpg'),
            AsyncMock(),
            AsyncMock(side_effect=PosterJob._ProcessingError('failed to upload')),
            Mock(),
            [
                call._obtain(),
                call._resize('poster.jpg', None, None),
                call._write('maybe_resized.jpg', ''),
                call._upload('maybe_resized.jpg', ()),
                call.error(PosterJob._ProcessingError('failed to upload')),
            ],
            id='Exception from _upload()',
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test_run(
        width, height, write_to, imghosts,
        obtain, resize, write, upload, get_poster_filename,
        exp_mock_calls,
        mocker,
):
    job = PosterJob(getter=AsyncMock(), width=width, height=height, write_to=write_to, imghosts=imghosts)
    mocks = AsyncMock()
    mocks.attach_mock(mocker.patch.object(job, '_obtain', obtain), '_obtain')
    mocks.attach_mock(mocker.patch.object(job, '_resize', resize), '_resize')
    mocks.attach_mock(mocker.patch.object(job, '_write', write), '_write')
    mocks.attach_mock(mocker.patch.object(job, '_upload', upload), '_upload')
    mocks.attach_mock(mocker.patch.object(job, '_get_poster_filename', get_poster_filename), '_get_poster_filename')
    mocks.attach_mock(mocker.patch.object(job, 'send'), 'send')
    mocks.attach_mock(mocker.patch.object(job, 'error'), 'error')

    await job.run()
    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize(
    argnames='params_or_poster, exp_result, exp_mock_calls',
    argvalues=(
        pytest.param(
            errors.RequestError('service down'),
            PosterJob._ProcessingError('Failed to get poster: service down'),
            [
                call.emit('obtaining'),
                call._getter(),
            ],
            id='Getter raises RequestError',
        ),
        pytest.param(
            '',
            PosterJob._ProcessingError('No poster found'),
            [
                call.emit('obtaining'),
                call._getter(),
            ],
            id='Getter returns empty string',
        ),
        pytest.param(
            None,
            PosterJob._ProcessingError('No poster found'),
            [
                call.emit('obtaining'),
                call._getter(),
            ],
            id='Getter returns None',
        ),
        pytest.param(
            'myposter.jpg',
            {'poster': 'myposter.jpg'},
            [
                call.emit('obtaining'),
                call._getter(),
                call.emit('obtained', 'myposter.jpg'),
            ],
            id='Getter returns file path or URL',
        ),
        pytest.param(
            {'poster': 'myposter.jpg', 'width': 123, 'foo': 'bar'},
            {'poster': 'myposter.jpg', 'width': 123, 'foo': 'bar'},
            [
                call.emit('obtaining'),
                call._getter(),
                call.emit('obtained', 'myposter.jpg'),
            ],
            id='Getter returns parameters',
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test__obtain(params_or_poster, exp_result, exp_mock_calls, job, mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(job.signal, 'emit'), 'emit')
    if isinstance(params_or_poster, Exception):
        mocks.attach_mock(mocker.patch.object(job, '_getter', side_effect=params_or_poster), '_getter')
    else:
        mocks.attach_mock(mocker.patch.object(job, '_getter', return_value=params_or_poster), '_getter')

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            await job._obtain()
    else:
        return_value = await job._obtain()
        assert return_value == exp_result
    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize(
    argnames=(
        'poster, width, height, cache_directory,'
        'get_poster_filepath, resize,'
        'exp_result, exp_mock_calls,'
    ),
    argvalues=(
        pytest.param(
            'path/to/myposter.jpg', None, None, 'path/to/cache',
            AsyncMock(return_value='filepath/myposter.jpg'),
            Mock(return_value='filepath/myposter.resized.jpg'),
            'path/to/myposter.jpg',
            [],
            id='No resizing',
        ),
        pytest.param(
            'path/to/myposter.jpg', 123, 456, 'path/to/cache',
            AsyncMock(return_value='filepath/myposter.jpg'),
            Mock(return_value='filepath/myposter.resized.jpg'),
            'filepath/myposter.resized.jpg',
            [
                call._get_poster_filepath('path/to/myposter.jpg'),
                call.emit('resizing', 'filepath/myposter.jpg'),
                call.resize(
                    'filepath/myposter.jpg',
                    target_directory='path/to/cache',
                    target_filename='myposter.123x456.jpg',
                    width=123,
                    height=456,
                ),
                call.emit('resized', 'filepath/myposter.resized.jpg'),
            ],
            id='Resizing successful',
        ),
        pytest.param(
            'path/to/myposter.jpg', 123, 456, 'path/to/cache',
            AsyncMock(return_value='filepath/myposter.jpg'),
            Mock(side_effect=errors.ImageResizeError('wat')),
            PosterJob._ProcessingError('Failed to resize poster: wat'),
            [
                call._get_poster_filepath('path/to/myposter.jpg'),
                call.emit('resizing', 'filepath/myposter.jpg'),
                call.resize(
                    'filepath/myposter.jpg',
                    target_directory='path/to/cache',
                    target_filename='myposter.123x456.jpg',
                    width=123,
                    height=456,
                ),
            ],
            id='Resizing failed',
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test__resize(
        poster, width, height, cache_directory,
        get_poster_filepath, resize,
        exp_result, exp_mock_calls,
        job, mocker,
):
    mocks = Mock()
    mocker.patch.object(type(job), 'cache_directory', PropertyMock(return_value=cache_directory))
    mocks.attach_mock(mocker.patch.object(job.signal, 'emit'), 'emit')
    mocks.attach_mock(mocker.patch.object(job, '_get_poster_filepath', get_poster_filepath), '_get_poster_filepath')
    mocks.attach_mock(mocker.patch('upsies.utils.image.resize', resize), 'resize')

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            await job._resize(poster, width, height)
    else:
        return_value = await job._resize(poster, width, height)
        assert return_value == exp_result
    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize(
    argnames=(
        'poster, filepath,'
        'read_file_or_url, write_file,'
        'exp_mock_calls,'
    ),
    argvalues=(
        pytest.param(
            'myposter.jpg', '',
            AsyncMock(return_value=b'image data'),
            AsyncMock(return_value='final/path/to/myposter.jpg'),
            [],
            id='Without destination file path',
        ),
        pytest.param(
            'myposter.jpg', 'path/to/myposter.jpg',
            AsyncMock(return_value=b'image data'),
            AsyncMock(return_value='final/path/to/myposter.jpg'),
            [
                call._read_file_or_url('myposter.jpg'),
                call._write_file(b'image data', 'path/to/myposter.jpg'),
                call.send('final/path/to/myposter.jpg'),
            ],
            id='With destination file path',
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test__write(
        poster, filepath,
        read_file_or_url, write_file,
        exp_mock_calls,
        job, mocker,
):
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(job, '_read_file_or_url', read_file_or_url), '_read_file_or_url')
    mocks.attach_mock(mocker.patch.object(job, '_write_file', write_file), '_write_file')
    mocks.attach_mock(mocker.patch.object(job, 'send'), 'send')

    await job._write(poster, filepath)
    assert mocks.mock_calls == exp_mock_calls


BAD_IMGHOST = Mock(upload=AsyncMock(side_effect=errors.RequestError('service is down')))
GOOD_IMGHOST = Mock(upload=AsyncMock(return_value='http://example.org/good/myposter.jpg'))
IRRELEVANT_IMGHOST = Mock(upload=AsyncMock(return_value='http://example.org/irrelevant/myposter.jpg'))

@pytest.mark.parametrize(
    argnames=(
        'poster, imghosts,'
        'get_poster_filepath,'
        'exp_mock_calls, exp_exception'
    ),
    argvalues=(
        pytest.param(
            'myposter.jpg',
            (),
            AsyncMock(return_value='path/to/myposter.jpg'),
            [],
            None,
            id='No image hosts',
        ),
        pytest.param(
            'myposter.jpg',
            (BAD_IMGHOST, GOOD_IMGHOST, IRRELEVANT_IMGHOST),
            AsyncMock(return_value='path/to/myposter.jpg'),
            [
                call._get_poster_filepath('myposter.jpg'),
                call.emit('uploading', BAD_IMGHOST),
                call.BAD_IMGHOST.upload('path/to/myposter.jpg', thumb_width=0),
                call.warn('Failed to upload poster: service is down'),
                call.emit('uploading', GOOD_IMGHOST),
                call.GOOD_IMGHOST.upload('path/to/myposter.jpg', thumb_width=0),
                call.emit('uploaded', 'http://example.org/good/myposter.jpg'),
                call.send('http://example.org/good/myposter.jpg'),
            ],
            None,
            id='Bad image host and good image host',
        ),
        pytest.param(
            'myposter.jpg',
            (BAD_IMGHOST, BAD_IMGHOST),
            AsyncMock(return_value='path/to/myposter.jpg'),
            [
                call._get_poster_filepath('myposter.jpg'),
                call.emit('uploading', BAD_IMGHOST),
                call.BAD_IMGHOST.upload('path/to/myposter.jpg', thumb_width=0),
                call.warn('Failed to upload poster: service is down'),
                call.emit('uploading', BAD_IMGHOST),
                call.BAD_IMGHOST.upload('path/to/myposter.jpg', thumb_width=0),
                call.warn('Failed to upload poster: service is down'),
            ],
            PosterJob._ProcessingError('All uploads failed'),
            id='Only bad image hosts',
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test__upload(
        poster, imghosts,
        get_poster_filepath,
        exp_mock_calls, exp_exception,
        job, mocker,
):
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(job, '_get_poster_filepath', get_poster_filepath), '_get_poster_filepath')
    mocks.attach_mock(BAD_IMGHOST, 'BAD_IMGHOST')
    mocks.attach_mock(GOOD_IMGHOST, 'GOOD_IMGHOST')
    mocks.attach_mock(IRRELEVANT_IMGHOST, 'IRRELEVANT_IMGHOST')
    mocks.attach_mock(mocker.patch.object(job.signal, 'emit'), 'emit')
    mocks.attach_mock(mocker.patch.object(job, 'send'), 'send')
    mocks.attach_mock(mocker.patch.object(job, 'warn'), 'warn')
    mocks.attach_mock(mocker.patch.object(job, 'error'), 'error')

    if isinstance(exp_exception, Exception):
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            await job._upload(poster, imghosts)
    else:
        await job._upload(poster, imghosts)
    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize(
    argnames=(
        'poster, read_file_or_url, cache_directory, get_poster_filename, write_file,'
        'exp_mock_calls, exp_return_value,'
    ),
    argvalues=(
        pytest.param(
            'http://my.poster.jpg',
            AsyncMock(return_value=b'image data'),
            'path/to/cache',
            Mock(return_value='poster_filename.jpg'),
            AsyncMock(return_value='path/to/poster.jpg'),
            [
                call._read_file_or_url('http://my.poster.jpg'),
                call._get_poster_filename('http://my.poster.jpg'),
                call._write_file(b'image data', 'path/to/cache/poster_filename.jpg'),
            ],
            'path/to/poster.jpg',
            id='Poster is URL',
        ),
        pytest.param(
            'myposter.jpg',
            AsyncMock(return_value=b'image data'),
            'path/to/cache',
            Mock(return_value='poster_filename.jpg'),
            AsyncMock(return_value='path/to/poster.jpg'),
            [],
            'myposter.jpg',
            id='Poster is file path',
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test__get_poster_filepath(
        poster, read_file_or_url, cache_directory, get_poster_filename, write_file,
        exp_mock_calls, exp_return_value,
        job, mocker,
):
    mocks = Mock()
    mocker.patch.object(type(job), 'cache_directory', PropertyMock(return_value=cache_directory))
    mocks.attach_mock(mocker.patch.object(job, '_read_file_or_url', read_file_or_url), '_read_file_or_url')
    mocks.attach_mock(mocker.patch.object(job, '_get_poster_filename', get_poster_filename), '_get_poster_filename')
    mocks.attach_mock(mocker.patch.object(job, '_write_file', write_file), '_write_file')

    # Return value should be cached.
    for _ in range(3):
        return_value = await job._get_poster_filepath(poster)
        assert return_value == exp_return_value
    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize(
    argnames='poster, exp_return_value',
    argvalues=(
        pytest.param(
            'http://example.org/path/to/poster.jpg?foo=bar',
            'poster:example.org.' + hashlib.md5(b'/path/to/poster.jpg.foo=bar').hexdigest() + '.jpg',
            id='Poster is URL with extension',
        ),
        pytest.param(
            'http://example.org/path/to/d34db33f?foo=bar',
            'poster:example.org.' + hashlib.md5(b'/path/to/d34db33f.foo=bar').hexdigest(),
            id='Poster is URL without extension',
        ),
        pytest.param(
            'path/to/poster.jpg',
            'poster.jpg',
            id='Poster is file path',
        ),
    ),
    ids=lambda v: repr(v),
)
def test__get_poster_filename(poster, exp_return_value, job):
    return_value = job._get_poster_filename(poster)
    assert return_value == exp_return_value


@pytest.mark.parametrize(
    argnames=(
        'poster, http_get,'
        'exp_result, exp_mock_calls,'
    ),
    argvalues=(
        pytest.param(
            'http://myposter.jpg',
            AsyncMock(return_value=Mock(bytes=b'image data')),
            b'image data',
            [
                call.emit('downloading', 'http://myposter.jpg'),
                call.http_get('http://myposter.jpg', cache=True),
                call.emit('downloaded', 'http://myposter.jpg'),
            ],
            id='Poster is good URL',
        ),
        pytest.param(
            'http://broken.host/myposter.jpg',
            AsyncMock(side_effect=errors.RequestError('service is down')),
            PosterJob._ProcessingError('Failed to download poster: service is down'),
            [
                call.emit('downloading', 'http://broken.host/myposter.jpg'),
                call.http_get('http://broken.host/myposter.jpg', cache=True),
            ],
            id='Poster is bad URL',
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test__read_file_or_url_gets_url(
        poster, http_get,
        exp_result, exp_mock_calls,
        job, mocker,
):
    mocks = Mock()
    mocks.attach_mock(mocker.patch('upsies.utils.http.get', http_get), 'http_get')
    mocks.attach_mock(mocker.patch.object(job.signal, 'emit'), 'emit')

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            await job._read_file_or_url(poster)
    else:
        return_value = await job._read_file_or_url(poster)
        assert return_value == exp_result
    assert mocks.mock_calls == exp_mock_calls

@pytest.mark.parametrize(
    argnames=(
        'poster, image_data,'
        'exp_result,'
    ),
    argvalues=(
        pytest.param(
            'path/to/myposter.jpg',
            b'image data',
            b'image data',
            id='Poster is good file',
        ),
        pytest.param(
            'path/to/myposter.jpg',
            None,
            PosterJob._ProcessingError('Failed to read poster: No such file or directory'),
            id='Poster is bad file',
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test__read_file_or_url_gets_filepath(
        poster, image_data,
        exp_result,
        job, tmp_path, mocker,
):
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(job.signal, 'emit'), 'emit')

    poster = tmp_path / 'poster.jpg'
    if image_data is not None:
        poster.write_bytes(image_data)

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            await job._read_file_or_url(str(poster))
    else:
        return_value = await job._read_file_or_url(str(poster))
        assert return_value == exp_result

    # No signals emitted when reading files.
    assert mocks.mock_calls == []


@pytest.mark.parametrize(
    argnames=(
        'data, filepath, sanitize_path, sanitized_path_is_writable,'
        'exp_result,'
    ),
    argvalues=(
        pytest.param(
            b'image data',
            '{tmp_path}/path/to/poster.jpg',
            Mock(side_effect=lambda path: f'{path}.sanitized'),
            True,
            '{tmp_path}/path/to/poster.jpg.sanitized',
            id='File is written',
        ),
        pytest.param(
            b'image data',
            '{tmp_path}/path/to/poster.jpg',
            Mock(side_effect=lambda path: f'{path}.sanitized'),
            False,
            PosterJob._ProcessingError('Failed to write {tmp_path}/path/to/poster.jpg.sanitized: Permission denied'),
            id='File cannot be written',
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test__write_file(
        data, filepath, sanitize_path, sanitized_path_is_writable,
        exp_result,
        job, tmp_path, mocker,
):
    mocks = Mock()
    mocks.attach_mock(mocker.patch('upsies.utils.fs.sanitize_path', sanitize_path), 'sanitize_path')
    mocks.attach_mock(mocker.patch.object(job.signal, 'emit'), 'emit')

    filepath = filepath.format(tmp_path=tmp_path)
    (tmp_path / filepath).parent.mkdir(parents=True, exist_ok=True)
    if not sanitized_path_is_writable:
        (tmp_path / filepath).parent.chmod(0o000)

    try:
        if isinstance(exp_result, Exception):
            msg = str(exp_result).format(tmp_path=tmp_path)
            with pytest.raises(type(exp_result), match=rf'^{re.escape(msg)}$'):
                await job._write_file(data, filepath)
        else:
            return_value = await job._write_file(data, filepath)
            assert return_value == exp_result.format(tmp_path=tmp_path)
    finally:
        (tmp_path / filepath).parent.chmod(0o700)

    assert mocks.mock_calls == [
        call.sanitize_path(filepath),
    ]



"""
Generic user dialogs

Prompts take arbitrary keyword arguments that are used by the user interface
to create the dialog.

When the user makes a choice, the user interface must pass it on to the prompt
object via :meth:`~.Prompt.set_result`. The prompt object then takes care of
passing the result to any callbacks.
"""

import asyncio

import logging  # isort:skip
_log = logging.getLogger(__name__)


class Prompt:
    """
    Base class of all prompts

    Every prompt takes at least a sequence of callbacks. Any other keyword
    arguments are specified by the subclass. The user interface can access
    them as :attr:`parameters`.
    """

    def __init__(self, *, callbacks=(), **parameters):
        self._callbacks = list(callbacks)
        self._parameters = parameters
        self._result_arrived = asyncio.Event()

    @property
    def parameters(self):
        """Keyword arguments from instantiation (except for `callbacks`)"""
        return self._parameters

    async def wait(self):
        """Block until :meth:`set_result` is called"""
        _log.debug('Waiting for %s ...', type(self).__name__)
        await self._result_arrived.wait()
        _log.debug('Waiting for %s', type(self).__name__)

    def on_result(self, callback):
        """Schedule `callback` to be called when :meth:`set_result` is called"""
        self._callbacks.append(callback)

    def set_result(self, result):
        """
        Take the result from the user dialog and pass it on to any previously
        scheduled callbacks
        """
        _log.debug('Setting result: %r', result)

        # Call callbacks
        for callback in self._callbacks:
            _log.debug('Reporting result to %r: %r', callback, result)
            callback(result)

        # Unblock any wait() calls
        self._result_arrived.set()

    def __repr__(self):
        arguments = []
        if self._parameters:
            arguments.append(', '.join((
                f'{k}={v!r}'
                for k, v in self._parameters.items()
            )))
        arguments.append(f'callbacks={self._callbacks}')
        arguments_string = ', '.join(arguments)
        return f'{type(self).__name__}({arguments_string})'


class RadioListPrompt(Prompt):
    """
    Pick one of two or more options

    :param options: Sequence of choices
    :param question: Question to show alongside the options or `None`
    """

    def __init__(self, *, callbacks, options, question=None):
        super().__init__(callbacks=callbacks, options=options, question=question)



import asyncio
from unittest.mock import Mock, call

import pytest

from upsies.uis import prompts


def test_Prompt_parameters():
    callbacks = (Mock(), Mock())
    prompt = prompts.Prompt(
        foo='bar',
        baz=123,
        callbacks=callbacks,
    )
    assert prompt.parameters == {
        'foo': 'bar',
        'baz': 123,
    }


@pytest.mark.asyncio
async def test_Prompt_set_result():
    callbacks = (Mock(), Mock())
    prompt = prompts.Prompt(
        foo='bar',
        baz=123,
        callbacks=callbacks,
    )

    assert not prompt._result_arrived.is_set()
    asyncio.get_event_loop().call_soon(prompt.set_result('the result'))
    await prompt.wait()
    assert prompt._result_arrived.is_set()
    for callback in callbacks:
        assert callback.call_args_list == [call('the result')]


def test_Prompt_on_result():
    callbacks = (Mock(), Mock())
    prompt = prompts.Prompt(
        foo='bar',
        baz=123,
        callbacks=callbacks,
    )
    additional_callback = Mock()
    prompt.on_result(additional_callback)
    assert prompt._callbacks == [
        callbacks[0],
        callbacks[1],
        additional_callback,
    ]


@pytest.mark.parametrize(
    argnames='callbacks, exp_callbacks_string',
    argvalues=(
        (
            (str, int),
            f'callbacks=[{str}, {int}]',
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.parametrize(
    argnames='parameters, exp_parameters_string',
    argvalues=(
        (
            {'foo': 'bar', 'baz': 123},
            "foo='bar', baz=123",
        ),
        (
            {},
            '',
        ),
    ),
    ids=lambda v: repr(v),
)
def test_Prompt___repr__(callbacks, exp_callbacks_string, parameters, exp_parameters_string):
    prompt = prompts.Prompt(
        callbacks=callbacks,
        **parameters,
    )
    repr_string = repr(prompt)
    exp_arguments_string = ', '.join(
        exp_arg
        for exp_arg in (exp_parameters_string, exp_callbacks_string)
        if exp_arg
    )
    assert repr_string == (
        f'{prompts.Prompt.__name__}('
        + exp_arguments_string
        + ')'
    )


@pytest.mark.parametrize('question', ('Hello?', None))
def test_RadioListPrompt(question):
    prompt = prompts.RadioListPrompt(
        options=('foo', 'bar', 'baz'),
        question=question,
        callbacks=(Mock(), Mock()),
    )

    assert prompt.parameters == {
        'options': ('foo', 'bar', 'baz'),
        'question': question,
    }



import functools

from . import JobWidgetBase


class MediainfoJobWidget(JobWidgetBase):

    is_interactive = False

    def setup(self):
        pass

    @functools.cached_property
    def runtime_widget(self):
        return None



import itertools
from unittest.mock import AsyncMock, Mock, PropertyMock, call

import pytest

from upsies import errors
from upsies.jobs.submit import SubmitJob
from upsies.trackers import TrackerBase, TrackerJobsBase


@pytest.fixture
def make_SubmitJob(tracker, tracker_jobs, tmp_path):
    def make_SubmitJob(**kwargs):
        return SubmitJob(
            home_directory=tmp_path,
            cache_directory=tmp_path,
            tracker=tracker,
            tracker_jobs=tracker_jobs,
            **kwargs,
        )

    return make_SubmitJob


@pytest.fixture
def job(make_SubmitJob):
    return make_SubmitJob()


@pytest.fixture
def tracker():
    class TestTracker(TrackerBase):
        name = 'test'
        label = 'TeST'
        TrackerConfig = 'tracker config class'
        TrackerJobs = 'tracker jobs class'
        login = AsyncMock()
        logout = AsyncMock()
        _login = AsyncMock()
        _logout = AsyncMock()
        is_logged_in = PropertyMock()
        get_announce_url = AsyncMock()
        upload = AsyncMock()

    return TestTracker()

@pytest.fixture
def tracker_jobs(tracker):
    kwargs = {
        'content_path': 'content/path',
        'tracker': tracker,
        'btclient': Mock(),
        'torrent_destination': 'torrent/destination',
        'image_hosts': (Mock(),),
        'common_job_args': {},
    }

    class TestTrackerJobs(TrackerJobsBase):
        def __init__(self, **kw):
            return super().__init__(**{**kwargs, **kw})

        jobs_before_upload = PropertyMock()
        jobs_after_upload = PropertyMock()
        submission_ok = PropertyMock(return_value=False)

    return TestTrackerJobs()


def test_cache_id(job):
    assert job.cache_id is None


@pytest.mark.asyncio
async def test_initialize_creates_enabled_jobs_after_upload(make_SubmitJob, mocker):
    logged = []

    def create_jobs(msg):
        logged.append(msg)

    mocker.patch('upsies.jobs.submit.SubmitJob.enabled_jobs_after_upload', PropertyMock(
        side_effect=lambda: create_jobs('Creating jobs_after_upload'),
    ))

    make_SubmitJob()

    assert logged == ['Creating jobs_after_upload']


@pytest.mark.asyncio
async def test_initialize_registers_to_tracker_and_tracker_jobs_signals(job):
    for t in (job._tracker, job._tracker_jobs):
        assert job.warn in t.signal.signals['warning']
        assert job.error in t.signal.signals['error']
        assert job.exception in t.signal.signals['exception']


@pytest.mark.asyncio
async def test_initialize_starts_enabled_jobs_after_upload_on_output(job, mocker):
    mocker.patch.object(type(job._tracker_jobs), 'jobs_after_upload', PropertyMock(
        return_value=[Mock(), Mock(), Mock()],
    ))

    assert len(job.enabled_jobs_after_upload) > 0
    for j in job.enabled_jobs_after_upload:
        assert j.start.call_args_list == []

    job.start()
    try:
        job.send('mock output')

        for j in job.enabled_jobs_after_upload:
            assert j.start.call_args_list == [call()]
    finally:
        await job.wait()


@pytest.mark.parametrize('signal', ('finished', 'error'))
def test_info_is_cleared(signal, job, mocker):
    job.info = 'foo'
    assert job.info == 'foo'

    job.signal.emit(signal, 'foo')

    assert job.info == ''


class MockJob:
    instances = {}

    def __init__(self, name, start, actions):
        self.name = name
        self._actions = list(actions)
        self._was_started = start
        self._is_finished = False
        self.wait_calls = 0
        type(self).instances[name] = self

    async def wait(self):
        self.wait_calls += 1
        print(f'=== {self.name}, {self.wait_calls=} ===============================')

        if self._actions:
            name, action = self._actions.pop(0).split('.')
            if action:
                attr, value = action.split('=')
                job = type(self).instances[name]
                setattr(job, attr, eval(value))

        for job in type(self).instances.values():
            print('   ', job)

    @property
    def was_started(self):
        return self._was_started

    @property
    def is_finished(self):
        return self._is_finished

    def __repr__(self):
        return f'<MockJob {self.name}, {self.was_started=}, {self.is_finished=})'


@pytest.mark.asyncio
async def test_run_keeps_waiting_for_jobs_until_all_are_finished(job, mocker):
    jobs = {
        'a': MockJob('a', start=True, actions=(
            '.',
            'b._was_started=True',
            'a._is_finished=True',
        )),

        'b': MockJob('b', start=False, actions=(
            '.',
            '.',
            'c._was_started=True',
            '.',
            '.',
            'b._is_finished=True',
        )),

        'c': MockJob('c', start=False, actions=(
            '.',
            '.',
            '.',
            'c._is_finished=True',
        )),
    }

    mocker.patch.object(type(job), 'enabled_jobs_before_upload', PropertyMock(side_effect=itertools.chain(
        # Only one job at first
        ((jobs['a'],),),
        # Job "b" was added
        ((jobs['a'], jobs['b']),),
        ((jobs['a'], jobs['b']),),
        # Job "c" was added
        itertools.repeat(
            (jobs['a'], jobs['b'], jobs['c']),
            100,
        ),
    )))

    await job.run()

    assert jobs['a'].wait_calls == 3
    assert jobs['b'].wait_calls == 6
    assert jobs['c'].wait_calls == 4


@pytest.mark.parametrize('submission_ok', ('submission_ok=True', 'submission_ok=False'))
@pytest.mark.asyncio
async def test_run_makes_calls_in_corret_order(submission_ok, job, mocker):
    jobs = {
        'a': MockJob('a', start=True, actions=(
            'a._is_finished=True',
        )),

        'b': MockJob('b', start=True, actions=(
            'b._is_finished=True',
        )),

        'c': MockJob('c', start=True, actions=(
            'c._is_finished=True',
        )),
    }

    mocker.patch.object(type(job), 'enabled_jobs_before_upload', PropertyMock(side_effect=itertools.chain(
        itertools.repeat(
            (jobs['a'], jobs['b'], jobs['c']),
            100,
        ),
    )))

    mocks = Mock()
    # Mock each job's wait() so we can record it, but still call the original
    # wait() by passing it as a side_effect.
    mocks.attach_mock(mocker.patch.object(jobs['a'], 'wait', AsyncMock(side_effect=jobs['a'].wait)), 'wait_a')
    mocks.attach_mock(mocker.patch.object(jobs['b'], 'wait', AsyncMock(side_effect=jobs['b'].wait)), 'wait_b')
    mocks.attach_mock(mocker.patch.object(jobs['c'], 'wait', AsyncMock(side_effect=jobs['c'].wait)), 'wait_c')
    mocks.attach_mock(mocker.patch.object(job._tracker, 'await_tasks'), 'await_tasks')
    mocks.attach_mock(mocker.patch.object(job, '_submit'), '_submit')
    mocker.patch.object(type(job._tracker_jobs), 'submission_ok', PropertyMock(return_value=submission_ok))

    await job.run()

    exp_mock_calls = [
        call.wait_a(),
        call.wait_b(),
        call.wait_c(),
        call.await_tasks(),
    ]
    if submission_ok:
        exp_mock_calls.append(call._submit())

    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize('torrent_page_url', ('http:/foo?torrent_id=123', ''))
@pytest.mark.parametrize('error_method', ('login', 'upload', 'logout'))
@pytest.mark.asyncio
async def test__submit_handles_RequestError_from_tracker_coro(error_method, torrent_page_url, job, mocker):
    mocks = Mock()
    mocker.patch.object(job._tracker, error_method, AsyncMock(
        side_effect=errors.RequestError(f'{error_method}: No connection'),
    ))
    if error_method != 'upload':
        job._tracker.upload.return_value = torrent_page_url

    mocks.attach_mock(getattr(job._tracker, 'login'), 'login')
    mocks.attach_mock(getattr(job._tracker, 'logout'), 'logout')
    mocks.attach_mock(getattr(job._tracker, 'upload'), 'upload')
    mocks.attach_mock(mocker.patch.object(job, 'send'), 'send')
    mocks.attach_mock(mocker.patch.object(job, 'error'), 'error')

    assert job.output == ()
    assert job.errors == ()

    await job._submit()

    exp_mock_calls = [
        call.login(),
    ]

    if error_method == 'login':
        exp_mock_calls.append(call.error(errors.RequestError(f'{error_method}: No connection')))

    elif error_method == 'upload':
        exp_mock_calls.append(call.upload(job._tracker_jobs))
        exp_mock_calls.append(call.logout())
        exp_mock_calls.append(call.error(errors.RequestError(f'{error_method}: No connection')))

    else:
        exp_mock_calls.append(call.upload(job._tracker_jobs))
        if torrent_page_url:
            exp_mock_calls.append(call.send(torrent_page_url))
        exp_mock_calls.append(call.logout())
        exp_mock_calls.append(call.error(errors.RequestError(f'{error_method}: No connection')))

    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.asyncio
async def test__submit_calls_methods_and_callbacks_in_correct_order(job, mocker):
    mocks = Mock()
    mocks.attach_mock(job._tracker.login, 'login')
    mocks.attach_mock(job._tracker.logout, 'logout')
    mocks.attach_mock(job._tracker.upload, 'upload')
    mocks.attach_mock(mocker.patch.object(job, 'send'), 'send')

    job.signal.register('logging_in', mocks.logging_in_cb)
    job.signal.register('logged_in', mocks.logged_in_cb)
    job.signal.register('uploading', mocks.uploading_cb)
    job.signal.register('uploaded', mocks.uploaded_cb)
    job.signal.register('logging_out', mocks.logging_out_cb)
    job.signal.register('logged_out', mocks.logged_out_cb)

    await job._submit()

    assert mocks.method_calls == [
        call.logging_in_cb(),
        call.login(),
        call.logged_in_cb(),

        call.uploading_cb(),
        call.upload(job._tracker_jobs),
        call.send(mocks.upload.return_value),
        call.uploaded_cb(),

        call.logging_out_cb(),
        call.logout(),
        call.logged_out_cb(),
    ]

@pytest.mark.parametrize(
    argnames='torrent_url, exp_send_calls',
    argvalues=(
        ('http://torrent.url', [call('http://torrent.url')]),
        ('', []),
        (None, []),
    ),
)
@pytest.mark.asyncio
async def test__submit_sends_upload_return_value_as_output(torrent_url, exp_send_calls, job, mocker):
    mocker.patch.object(job, 'send')
    mocker.patch.object(job._tracker, 'upload', return_value=torrent_url)
    await job._submit()
    assert job.send.call_args_list == exp_send_calls


@pytest.mark.asyncio
async def test__submit_sets_info_property_to_current_status(job, mocker):
    exp_infos = [
        'Logging in',
        'Logged in',
        'Uploading',
        'Uploaded',
        'Logging out',
        '',
    ]

    def info_cb():
        assert job.info == exp_infos.pop(0)

    job.signal.register('logging_in', info_cb)
    job.signal.register('logged_in', info_cb)
    job.signal.register('uploading', info_cb)
    job.signal.register('uploaded', info_cb)
    job.signal.register('logging_out', info_cb)
    job.signal.register('logged_out', info_cb)

    mocker.patch.object(job, 'send')
    await job._submit()

    assert exp_infos == []


def test_enabled_jobs_before_upload(job, mocker):
    jobs_before_upload = {
        'a': None,
        'b': Mock(is_enabled=False),
        'c': None,
        'd': Mock(is_enabled=True),
        'e': None,
        'f': Mock(is_enabled=False),
        'g': None,
        'h': Mock(is_enabled=True),
    }
    mocker.patch.object(type(job._tracker_jobs), 'jobs_before_upload', PropertyMock(return_value=jobs_before_upload.values()))

    assert job.enabled_jobs_before_upload == (
        jobs_before_upload['d'],
        jobs_before_upload['h'],
    )


def test_enabled_jobs_after_upload(job, mocker):
    jobs_after_upload = {
        'a': None,
        'b': Mock(is_enabled=False),
        'c': None,
        'd': Mock(is_enabled=True),
        'e': None,
        'f': Mock(is_enabled=False),
        'g': None,
        'h': Mock(is_enabled=True),
    }
    mocker.patch.object(type(job._tracker_jobs), 'jobs_after_upload', PropertyMock(return_value=jobs_after_upload.values()))

    assert job.enabled_jobs_after_upload == (
        jobs_after_upload['d'],
        jobs_after_upload['h'],
    )


@pytest.mark.parametrize('submission_ok, exp_hidden', ((False, True), (True, False)))
def test_hidden_property(submission_ok, exp_hidden, job, mocker):
    mocker.patch.object(type(job._tracker_jobs), 'submission_ok', PropertyMock(return_value=submission_ok))
    assert job.hidden is exp_hidden


@pytest.mark.parametrize(
    argnames='submission_ok, jobs, exp_final_job',
    argvalues=(
        (True, (), None),
        (False, (Mock(is_finished=True), Mock(is_finished=False), Mock(is_finished=True)), None),
        (False, (Mock(is_finished=True, id='a'), Mock(is_finished=True, id='b'), Mock(is_finished=True, id='c')), 'c'),
    ),
)
def test_final_job_before_upload_property(submission_ok, jobs, exp_final_job, job, mocker):
    mocker.patch.object(type(job._tracker_jobs), 'submission_ok', PropertyMock(return_value=submission_ok))
    mocker.patch.object(type(job), 'enabled_jobs_before_upload', PropertyMock(return_value=jobs))
    if not exp_final_job:
        assert job.final_job_before_upload is None
    else:
        assert job.final_job_before_upload.id == exp_final_job


@pytest.mark.parametrize(
    argnames='own_output, final_job_before_upload, exp_output',
    argvalues=(
        (('torrent url',), None, ('torrent url',)),
        (('torrent url',), Mock(output=('description',)), ('description',)),
    ),
    ids=lambda v: repr(v),
)
def test_output_property(own_output, final_job_before_upload, exp_output, job, mocker):
    parent_class = type(job).mro()[1]
    mocker.patch.object(parent_class, 'output', PropertyMock(return_value=own_output))
    mocker.patch.object(type(job), 'final_job_before_upload', PropertyMock(return_value=final_job_before_upload))
    assert job.output == exp_output


@pytest.mark.parametrize(
    argnames='own_exit_code, final_job_before_upload, exp_exit_code',
    argvalues=(
        (123, None, 123),
        (123, Mock(exit_code=456), 456),
    ),
)
def test_exit_code_property(own_exit_code, final_job_before_upload, exp_exit_code, job, mocker):
    parent_class = type(job).mro()[1]
    mocker.patch.object(parent_class, 'exit_code', PropertyMock(return_value=own_exit_code))
    mocker.patch.object(type(job), 'final_job_before_upload', PropertyMock(return_value=final_job_before_upload))
    assert job.exit_code == exp_exit_code



import copy
import hashlib
import random
import re
import sys
from unittest.mock import AsyncMock, Mock, PropertyMock, call

import pytest

from upsies import __project_name__, errors
from upsies.utils import imghosts


def make_TestImageHost(default_config=None, **kwargs):
    # Avoid NameError bug (https://github.com/python/cpython/issues/87546)
    default_config_ = default_config

    class TestImageHost(imghosts.ImageHostBase):
        name = 'imgw00t'
        default_config = default_config_ or {}

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._upload_image_mock = AsyncMock()

        async def _upload_image(self, image_path):
            return await self._upload_image_mock(image_path)

    return TestImageHost(**kwargs)


def test_cache_directory_property(mocker, tmp_path):
    mocker.patch('upsies.constants.DEFAULT_CACHE_DIRECTORY', 'mock/cache/path')
    imghost = make_TestImageHost()
    assert imghost.cache_directory == 'mock/cache/path'
    imghost = make_TestImageHost(cache_directory=tmp_path)
    assert imghost.cache_directory is tmp_path
    imghost.cache_directory = 'path/to/foo'
    assert imghost.cache_directory == 'path/to/foo'


def test_default_config_without_custom_defaults():
    imghost = make_TestImageHost()
    default_config_common = copy.deepcopy(imghost.default_config_common)
    exp_default_config = {**default_config_common, **imghost.default_config}
    assert imghost.default_config == exp_default_config
    exp_options = {**imghost.default_config}
    assert imghost.options == exp_options
    assert imghost.default_config_common == default_config_common

def test_default_config_with_custom_defaults():
    imghost = make_TestImageHost(default_config={'foo': 1, 'bar': 2})
    default_config_common = copy.deepcopy(imghost.default_config_common)
    exp_default_config = {**default_config_common, **{'foo': 1, 'bar': 2}}
    assert imghost.default_config == exp_default_config
    exp_options = exp_default_config
    assert imghost.options == exp_options
    assert imghost.default_config_common == default_config_common

def test_default_config_with_custom_defaults_and_custom_options():
    imghost = make_TestImageHost(default_config={'foo': 1, 'bar': 2}, options={'foo': 1, 'bar': 99})
    default_config_common = copy.deepcopy(imghost.default_config_common)
    exp_default_config = {**default_config_common, **{'foo': 1, 'bar': 2}}
    assert imghost.default_config == exp_default_config
    exp_options = {**imghost.default_config, **{'foo': 1, 'bar': 99}}
    assert imghost.options == exp_options
    assert imghost.default_config_common == default_config_common

def test_default_config_with_overloaded_common_defaults():
    default_config_key = tuple(imghosts.ImageHostBase.default_config_common)[0]
    default_config = {default_config_key: 'foozbar'}
    imghost = make_TestImageHost(default_config=default_config)
    default_config_common = copy.deepcopy(imghost.default_config_common)
    exp_default_config = {**default_config_common, **{default_config_key: 'foozbar'}}
    assert imghost.default_config == exp_default_config
    exp_options = exp_default_config
    assert imghost.options == exp_options
    assert imghost.default_config_common == default_config_common


def test_options_property():
    imghost = make_TestImageHost()
    assert imghost.options == {'thumb_width': 0}
    imghost = make_TestImageHost(default_config={'foo': 1, 'bar': 2})
    assert imghost.options == {'thumb_width': 0, 'foo': 1, 'bar': 2}
    imghost = make_TestImageHost(default_config={'foo': 1, 'bar': 2}, options={'bar': 99})
    assert imghost.options == {'thumb_width': 0, 'foo': 1, 'bar': 99}


def test_description():
    imghost = make_TestImageHost()
    assert imghost.description == ''


@pytest.mark.parametrize(
    argnames='options, exp_exception',
    argvalues=(
        ({}, None),
        ({'apikey': 'd34db33f'}, None),
        ({'apikey': ''}, errors.RequestError(
            'You must configure an API key first. Run '
            f'"{__project_name__} upload-images {{name}} --help" '
            'for more information.'
        )),
    ),
)
@pytest.mark.asyncio
async def test_upload_checks_for_missing_apikey(options, exp_exception, mocker, tmp_path):
    resize_mock = mocker.patch('upsies.utils.image.resize')
    ih = make_TestImageHost(cache_directory=tmp_path, options=options)
    mocker.patch.object(ih, '_get_image_url', AsyncMock())
    if exp_exception is None:
        await ih.upload('foo.png')
        assert resize_mock.call_args_list == []
        assert ih._get_image_url.call_args_list == [call('foo.png', cache=True)]
    else:
        exp_error = str(exp_exception).format(name=ih.name)
        with pytest.raises(type(exp_exception), match=rf'{re.escape(exp_error)}$'):
            await ih.upload('foo.png')
        assert resize_mock.call_args_list == []
        assert ih._get_image_url.call_args_list == []


@pytest.mark.parametrize('cache', (True, False), ids=('cache=True', 'cache=False'))
@pytest.mark.parametrize('thumb_width', (0, 123), ids=lambda v: f'thumb_width={v}')
@pytest.mark.asyncio
async def test_upload_succeeds(thumb_width, cache, mocker, tmp_path):
    resize_mock = mocker.patch('upsies.utils.image.resize', return_value='thumbnail.png')

    ih = make_TestImageHost(cache_directory=tmp_path, options={'thumb_width': thumb_width})
    image_urls = ['https://localhost:123/foo.png']
    if thumb_width:
        image_urls.append('https://localhost:123/foo.thumb.png')
    mocker.patch.object(ih, '_get_image_url', AsyncMock(side_effect=image_urls))

    image = await ih.upload('path/to/foo.png', cache=cache)

    if thumb_width:
        assert resize_mock.call_args_list == [call(
            'path/to/foo.png',
            width=thumb_width,
            target_directory=ih.cache_directory,
            overwrite=not cache,
        )]
        assert ih._get_image_url.call_args_list == [
            call('path/to/foo.png', cache=cache),
            call(resize_mock.return_value, cache=cache),
        ]
    else:
        assert ih._get_image_url.call_args_list == [
            call('path/to/foo.png', cache=cache),
        ]

    assert isinstance(image, imghosts.UploadedImage)
    assert image == image_urls[0]
    if thumb_width:
        assert image.thumbnail_url == image_urls[1]
    else:
        assert image.thumbnail_url is None

@pytest.mark.parametrize(
    argnames='cache, resize_error, exp_request_error',
    argvalues=(
        (False, errors.ImageResizeError('not an image'), errors.RequestError('not an image')),
        (True, errors.ImageResizeError('not an image'), errors.RequestError('not an image')),
    ),
)
@pytest.mark.asyncio
async def test_upload_catches_ResizeError(cache, resize_error, exp_request_error, mocker, tmp_path):
    resize_mock = mocker.patch('upsies.utils.image.resize', side_effect=resize_error)

    ih = make_TestImageHost(cache_directory=tmp_path, options={'thumb_width': 345})
    image_urls = ['https://localhost:123/foo.png', 'https://localhost:123/foo.thumb.png']
    mocker.patch.object(ih, '_get_image_url', AsyncMock(side_effect=image_urls))

    with pytest.raises(type(exp_request_error), match=rf'^{re.escape(str(exp_request_error))}$'):
        await ih.upload('path/to/foo.png', cache=cache)

    assert resize_mock.call_args_list == [call(
        'path/to/foo.png',
        width=345,
        target_directory=ih.cache_directory,
        overwrite=not cache,
    )]
    assert ih._get_image_url.call_args_list == [
        call('path/to/foo.png', cache=cache),
    ]


@pytest.mark.asyncio
async def test_get_image_url_from_cache(mocker, tmp_path):
    ih = make_TestImageHost(cache_directory=tmp_path)
    mocker.patch.object(ih, '_upload_image', AsyncMock(return_value='http://localhost:123/uploaded.image.jpg'))
    mocker.patch.object(ih, '_get_url_from_cache', Mock(return_value='http://localhost:123/cached.image.jpg'))
    mocker.patch.object(ih, '_store_url_to_cache')
    url = await ih._get_image_url('path/to/image.jpg', cache=True)
    assert url == 'http://localhost:123/cached.image.jpg'
    assert ih._get_url_from_cache.call_args_list == [call('path/to/image.jpg')]
    assert ih._upload_image.call_args_list == []
    assert ih._store_url_to_cache.call_args_list == []

@pytest.mark.asyncio
async def test_get_image_url_from_upload(mocker, tmp_path):
    ih = make_TestImageHost(cache_directory=tmp_path)
    mocker.patch.object(ih, '_upload_image', AsyncMock(return_value='http://localhost:123/uploaded.image.jpg'))
    mocker.patch.object(ih, '_get_url_from_cache', Mock(return_value='http://localhost:123/cached.image.jpg'))
    mocker.patch.object(ih, '_store_url_to_cache')
    url = await ih._get_image_url('path/to/image.jpg', cache=False)
    assert url == 'http://localhost:123/uploaded.image.jpg'
    assert ih._get_url_from_cache.call_args_list == []
    assert ih._upload_image.call_args_list == [call('path/to/image.jpg')]
    assert ih._store_url_to_cache.call_args_list == [call(
        'path/to/image.jpg',
        'http://localhost:123/uploaded.image.jpg',
    )]


def test__get_url_from_cache_finds_no_cache_file(tmp_path, mocker):
    ih = make_TestImageHost(cache_directory=tmp_path)
    mocker.patch.object(ih, '_get_cache_file_suffix', return_value='d34db33f.foohost.url')
    image_path = 'some/path/to/foo.png'

    url = ih._get_url_from_cache(image_path)
    assert url is None
    assert ih._get_cache_file_suffix.call_args_list == [call(image_path)]

def test__get_url_from_cache_cannot_read_cache_file(tmp_path, mocker):
    ih = make_TestImageHost(cache_directory=tmp_path)
    image_path = 'some/path/to/foo.png'

    mocker.patch.object(ih, '_get_cache_file_suffix', return_value='d34db33f.foohost.url')
    cache_file = tmp_path / f'ANYTHING_1.{ih._get_cache_file_suffix.return_value}'
    cache_file.write_text(' http://1.cached.url ')
    cache_file.chmod(0o000)

    try:
        url = ih._get_url_from_cache(image_path)
    finally:
        cache_file.chmod(0o600)

    assert url is None
    assert ih._get_cache_file_suffix.call_args_list == [call(image_path)]

def test__get_url_from_cache_with_cache_directory_that_contains_invalid_glob_pattern(tmp_path, mocker):
    cache_directory = tmp_path / '[Dual-Audio]'
    cache_directory.mkdir()
    ih = make_TestImageHost(cache_directory=cache_directory)
    image_path = 'some/path/to/foo.png'

    mocker.patch.object(ih, '_get_cache_file_suffix', return_value='d34db33f.foohost.url')
    (cache_directory / f'ANYTHING.{ih._get_cache_file_suffix.return_value}').write_text('http://1.cached.url')

    url = ih._get_url_from_cache(image_path)
    assert url in (
        'http://1.cached.url',
    )
    assert ih._get_cache_file_suffix.call_args_list == [call(image_path)]

def test__get_url_from_cache_succeeds(tmp_path, mocker):
    ih = make_TestImageHost(cache_directory=tmp_path)
    image_path = 'some/path/to/foo.png'

    mocker.patch.object(ih, '_get_cache_file_suffix', return_value='d34db33f.foohost.url')
    (tmp_path / f'ANYTHING_1.{ih._get_cache_file_suffix.return_value}').write_text(' http://1.cached.url ')
    (tmp_path / f'ANYTHING_2.{ih._get_cache_file_suffix.return_value}').write_text('  http://2.cached.url  ')

    url = ih._get_url_from_cache(image_path)
    assert url in (
        'http://1.cached.url',
        'http://2.cached.url',
    )
    assert ih._get_cache_file_suffix.call_args_list == [call(image_path)]


@pytest.mark.parametrize(
    argnames='image_filename, cache_file_suffix, exp_cache_filename',
    argvalues=(
        pytest.param(
            'foo.png',
            'd34db33f.foohost.url',
            'foo.png.d34db33f.foohost.url',
            id='Short file name',
        ),
        pytest.param(
            'foo.' + ('o' * 300) + '.png',
            'd34db33f.foohost.url',
            ('foo.' + ('o' * 300) + '.png.d34db33f.foohost.url')[-(250 - len('d34db33f.foohost.url')):],
            id='Long file name',
        ),
    ),
)
def test__store_url_to_cache_succeeds(image_filename, cache_file_suffix, exp_cache_filename, tmp_path, mocker):
    ih = make_TestImageHost(cache_directory=tmp_path)
    image_path = f'some/path/to/{image_filename}'
    url = 'http://image.url'

    mocker.patch.object(type(ih), 'cache_directory', PropertyMock(return_value=tmp_path))
    mocker.patch.object(ih, '_get_cache_file_suffix', return_value=cache_file_suffix)

    ih._store_url_to_cache(image_path, url)

    exp_cache_filepath = tmp_path / exp_cache_filename
    print(exp_cache_filepath)
    assert exp_cache_filepath.read_text() == url
    assert ih._get_cache_file_suffix.call_args_list == [call(image_path)]

def test__store_url_to_cache_cannot_create_cache_directory(tmp_path, mocker):
    ih = make_TestImageHost(cache_directory=tmp_path)
    image_path = 'some/path/to/image.png'
    url = 'http://image.url'
    cache_directory = tmp_path / 'parent' / 'cache'
    cache_directory.parent.mkdir()
    cache_directory.parent.chmod(0o000)

    try:
        mocker.patch.object(type(ih), 'cache_directory', PropertyMock(return_value=str(cache_directory)))
        mocker.patch.object(ih, '_get_cache_file_suffix', return_value='d34db33f.imghost.url')

        exp_cache_file = cache_directory / 'image.png.d34db33f.imghost.url'
        exp_msg = f'Unable to write cache {exp_cache_file}: {cache_directory}: Permission denied'
        with pytest.raises(RuntimeError, match=rf'^{re.escape(exp_msg)}$'):
            ih._store_url_to_cache(image_path, url)
    finally:
        cache_directory.parent.chmod(0o700)

def test__store_url_to_cache_cannot_create_cache_file(tmp_path, mocker):
    ih = make_TestImageHost(cache_directory=tmp_path)
    image_path = 'some/path/to/image.png'
    url = 'http://image.url'
    cache_directory = tmp_path / 'cache'
    cache_directory.mkdir()
    cache_file = cache_directory / 'image.png.d34db33f.imghost.url'
    cache_file.write_text('http://secret.url')
    cache_file.chmod(0o000)

    try:
        mocker.patch.object(type(ih), 'cache_directory', PropertyMock(return_value=str(cache_directory)))
        mocker.patch.object(ih, '_get_cache_file_suffix', return_value='d34db33f.imghost.url')

        exp_cache_file = cache_directory / 'image.png.d34db33f.imghost.url'
        exp_msg = f'Unable to write cache {exp_cache_file}: Permission denied'
        with pytest.raises(RuntimeError, match=rf'^{re.escape(exp_msg)}$'):
            ih._store_url_to_cache(image_path, url)
    finally:
        cache_file.chmod(0o600)


@pytest.mark.skipif(sys.version_info <= (3, 9, 0), reason='Python <= 3.8 is lacking random.randbytes')
def test__get_cache_file_suffix_gets_unique_id_from_file_content(tmp_path, mocker):
    ih = make_TestImageHost(cache_directory=tmp_path)
    image_data = random.randbytes(11 * 1024)
    image_path = tmp_path / 'image.png'
    image_path.write_bytes(image_data)

    cache_file_suffix = ih._get_cache_file_suffix(image_path)
    assert cache_file_suffix == (
        hashlib.md5(image_data[:10 * 1024]).hexdigest()
        + f'.{ih.name}.url'
    )

def test__get_cache_file_suffix_gets_unique_id_from_image_path(tmp_path, mocker):
    ih = make_TestImageHost(cache_directory=tmp_path)
    image_path = tmp_path / 'image.png'
    image_path.write_bytes(b'whatever')
    image_path.chmod(0o000)
    try:
        cache_file_suffix = ih._get_cache_file_suffix(str(image_path))
        assert cache_file_suffix == (
            hashlib.md5(str(image_path).encode('utf8')).hexdigest()
            + f'.{ih.name}.url'
        )
    finally:
        image_path.chmod(0o600)
